//================================================================================================
// $Id: cgvisitor.cpp 6465 2023-02-27 20:07:26Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator visitor
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "cgvisitor.h"
#include "codegen.h"
#include "cgsection.h"
#include "parserext.h"
#include "framestack.h"
#include "validator.h"
#include "wtruntime/wtruntime.h"
#include "wtruntime/types.h"
#include "wtruntime/stdlib.h"
#include <set>
#include <deque>
#include <sstream>
#include "wtparser/checker.h"
#include "wtparser/astvisitor.h"
#include "wtparser/ast/ast_program.h"
#include "wtparser/ast/ast_entity.h"
#include "wtparser/ast/ast_metadata.h"
#include "wtparser/ast/ast_block.h"
#include "wtparser/ast/ast_mainblock.h"
#include "wtparser/ast/ast_constructorblock.h"
#include "wtparser/ast/ast_destructorblock.h"
#include "wtparser/ast/ast_declareblock.h"
#include "wtparser/ast/ast_compound.h"
#include "wtparser/ast/ast_exceptioncompound.h"
#include "wtparser/ast/ast_ehandler.h"
#include "wtparser/ast/ast_ehandlers.h"
#include "wtparser/ast/ast_finally.h"
#include "wtparser/ast/ast_statement.h"
#include "wtparser/ast/ast_variableaccess.h"
#include "wtparser/ast/ast_variabledeclaration.h"
#include "wtparser/ast/ast_variabledeclfragment.h"
#include "wtparser/ast/ast_elementaccess.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_typeinfo.h"
#include "wtparser/ast/ast_simultaneoustype.h"
#include "wtparser/ast/ast_basictype.h"
#include "wtparser/ast/ast_seqtype.h"
#include "wtparser/ast/ast_simultaneousexpression.h"
#include "wtparser/ast/ast_stringliteral.h"
#include "wtparser/ast/ast_booleanliteral.h"
#include "wtparser/ast/ast_characterliteral.h"
#include "wtparser/ast/ast_numberliteral.h"
#include "wtparser/ast/ast_assignment.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"
#include "wtparser/ast/ast_functiondefinition.h"
#include "wtparser/ast/ast_functiondeclaration.h"
#include "wtparser/ast/ast_functioncall.h"
#include "wtparser/ast/ast_procedurecall.h"
#include "wtparser/ast/ast_recordfield.h"
#include "wtparser/ast/ast_recordtype.h"
#include "wtparser/ast/ast_recorddeclaration.h"
#include "wtparser/ast/ast_exceptiondeclaration.h"
#include "wtparser/ast/ast_precondition.h"
#include "wtparser/ast/ast_postcondition.h"
#include "wtparser/ast/ast_memberaccess.h"
#include "wtparser/ast/ast_andthen.h"
#include "wtparser/ast/ast_orelse.h"
#include "wtparser/ast/ast_returnstatement.h"
#include "wtparser/ast/ast_operatordeclaration.h"
#include "wtparser/ast/ast_raisestatement.h"
#include "wtparser/ast/ast_whilestatement.h"
#include "wtparser/ast/ast_dowhilestatement.h"
#include "wtparser/ast/ast_forstatement.h"
#include "wtparser/ast/ast_foreachstatement.h"
#include "wtparser/ast/ast_ifstatement.h"
#include "wtparser/ast/ast_assertstatement.h"
#include "wtparser/ast/ast_seqliteral.h"
#include "wtparser/ast/ast_recordliteral.h"
#include "wtparser/ast/ast_fieldassign.h"
#include "wtparser/ast/ast_datatype.h"
#include "wtparser/ast/ast_datatypecons.h"
#include "wtparser/ast/ast_datatypedeclaration.h"
#include "wtparser/ast/ast_datatypeliteral.h"
#include "wtparser/ast/ast_switchstatement.h"
#include "wtparser/ast/ast_switchcase.h"
#include "wtparser/ast/ast_datatypecasestatement.h"
#include "wtparser/ast/ast_defaultstatement.h"
#include "wtparser/ast/ast_functiontype.h"
#include "wtparser/ast/ast_functiontypedeclaration.h"
#include "wtparser/ast/ast_functionpointercall.h"
#include "wtparser/ast/ast_makefunctionpointer.h"
#include "wtparser/ast/ast_unresolvedfunctionname.h"
#include "wtparser/ast/ast_enumtype.h"
#include "wtparser/ast/ast_enumtypedeclaration.h"
#include "wtparser/ast/ast_enumerator.h"
#include "wtparser/ast/ast_enumcasestatement.h"
#include "wtparser/ast/ast_enumliteral.h"
#include "wtparser/ast/ast_recordupdate.h"

namespace wtc
{
namespace codegen
{

// TODO
//class VisitorBase : public wtast::AstVisitor
//{
//	CodeGen &          mCodeGen;
//	InstrumentEngine & mEngine;
//	FrameStack         mFrameStack;

class GenVisitor : public wtast::AstVisitor
{
	CodeGen &                  mCodeGen;
	InstrumentEngine &         mEngine;
	DataSection*               mDataSection;
	ExternalsSection*          mExternals;
	FrameStack                 mFrameStack;
	std::size_t                mInstrCount;
	std::string                mLabelCleanup;
	std::string                mLabelFinally;
	std::string                mLabelExhan;
	std::string                mLabelAftereh;
	VisitMode                  mVisitMode;
	bool                       mInGlobalBlock;
	wtast::FunctionDefinition* mFuncDef;
	wtast::PostCondition*      mPostCond;

	void EnterFrame(const std::string & name)
	{
 		if(mFrameStack.NoMoreFrames())
		{
			mInstrCount = mEngine.GetInstrumentCount();
		}

		mFrameStack.EnterFrame(name);
	}

	void GenCSOSDestructorCode()
	{
		const FrameStack::varlist_t & objlist = mFrameStack.GetObjectsInFrame();

		for(FrameStack::varlist_t::const_reverse_iterator it=objlist.rbegin(),end=objlist.rend();it!=end;++it)
		{
			mEngine.AddSingle(WT_TOMB_OP_OSSP);
			mEngine.AddSingle(WT_TOMB_OP_JF_I, 4);
			mEngine.AddSingle(WT_TOMB_OP_OSPO);
			mEngine.AddSingle(WT_TOMB_OP_IDRD_I, sizeof(wt_pfn));
			mEngine.AddSingle(WT_TOMB_OP_CFP);
		}
	}

	void LeaveFrame()
	{
		if(mPostCond && mFrameStack.IsLastFrame())
		{
			AddPostCondFooter(mPostCond);
			mPostCond = 0;
		}

		GenCSOSDestructorCode();

		mFrameStack.LeaveFrame();

 		if(mFrameStack.NoMoreFrames())
		{
			wt_u32 dvt_size  = mFrameStack.GetDvtSize();
			wt_u32 csos_size = mFrameStack.GetCsosSize();

			if(0!=dvt_size)
			{
				if(0==(dvt_size & 0xFFFFFF00ul))
				{
					mEngine.InsertSingle(mInstrCount, WT_TOMB_OP_DVTA_I, wt_u8(dvt_size));
				}
				else
				if(0==(dvt_size & 0xFF000000ul))
				{
					mEngine.InsertDouble(mInstrCount, WT_TOMB_OP_DVTA_L, wt_u8(dvt_size>>16), wt_u16(dvt_size));
				}
				else
				{
					mEngine.InsertDouble(mInstrCount++, WT_TOMB_OP_PWZ_L, 0, wt_u16(dvt_size>>16));
					mEngine.InsertDouble(mInstrCount++, WT_TOMB_OP_LSOR_L, 0, wt_u16(dvt_size));
					mEngine.InsertSingle(mInstrCount, WT_TOMB_OP_DVTA_S);
				}

				++mInstrCount;
			}

			if(0!=csos_size)
			{
				if(0==(csos_size & 0xFFFFFF00ul))
				{
					mEngine.InsertSingle(mInstrCount, WT_TOMB_OP_OSA_I, wt_u8(csos_size));
				}
				else
				if(0==(csos_size & 0xFF000000ul))
				{
					mEngine.InsertDouble(mInstrCount, WT_TOMB_OP_OSA_L, wt_u8(csos_size>>16), wt_u16(csos_size));
				}
				else
				{
					mEngine.InsertDouble(mInstrCount++, WT_TOMB_OP_PWZ_L, 0, wt_u16(csos_size>>16));
					mEngine.InsertDouble(mInstrCount++, WT_TOMB_OP_LSOR_L, 0, wt_u16(csos_size));
					mEngine.InsertSingle(mInstrCount, WT_TOMB_OP_OSA_S);
				}

				++mInstrCount;
			}

			if(0!=csos_size)
			{
				mEngine.AddSingle(WT_TOMB_OP_OSF);
			}

			if(0!=dvt_size)
			{
				mEngine.AddSingle(WT_TOMB_OP_DVTF);
			}

			mEngine.AddSingle(WT_TOMB_OP_LEAVE);
			mLabelFinally.clear();
		}
	}

	void CallFinallyIfNeeded()
	{
		if(!mLabelFinally.empty())
		{
			mEngine.AddSingle(WT_TOMB_OP_PPC);
			mEngine.AddSingle(WT_TOMB_OP_ADD_I, 3); // 3: add.i, jmp.l
			mEngine.AddJmpL(mLabelFinally);
		}
	}

	bool HaveEHandlers(wtast::ExceptionCompound* ec)
	{
		return ec->GetEHandlers() && ec->GetEHandlers()->GetNumberOfEHandlers();
	}

	bool HaveFinally(wtast::ExceptionCompound* ec)
	{
		return ec->GetFinally() && ec->GetFinally()->GetBody();
	}

	void GenPushVarAddr(const FrameStack::Variable* var)
	{
		if(var->is_global)
		{
			mEngine.MakePBA(mCodeGen.GetGlobalAddress() + var->offset);
		}
		else
		{
			mEngine.MakePVT(var->offset);
		}
	}

	void CopyReturnValueIfNeeded(wtast::Expression* expr, wt_u32 offs)
	{
		wtast::Type* type = static_cast<wtast::Type*>(expr->GetType());

		if(type->IsObject() && !expr->IsTemporary())
		{
			if(0==offs)
			{
				mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
			}
			else
			{
				mEngine.MakePDSP(offs);
				mEngine.AddSingle(WT_TOMB_OP_IDRD_I);
			}

			mCodeGen.GenCopyCall(type);
			mEngine.MakePDSP(offs + sizeof(wt_vmptr));
			mEngine.MakeST(sizeof(wt_vmptr));
		}
	}

	const FrameStack::Variable* FindVarInFormals(const std::string & name)
	{
		if(AddAssert(mCodeGen.mProblemHandler, mFuncDef))
		{
			return 0;
		}

		wtast::FunctionSignature* sig = mFuncDef->GetFunctionSignature();

		for(std::size_t i=0,n=sig->GetArgc();i!=n;++i)
		{
			wtast::FunctionArgument* arg = sig->GetArg(i);

			if(name==arg->GetName())
			{
				return mFrameStack.FindVar(arg->GetUniqueId());
			}
		}

		return 0;
	}

	void GenST(wtast::Type* type)
	{
		if(type->IsFunctionType())
		{
			mEngine.AddSingle(WT_TOMB_OP_STFP);
		}
		else
		{
			GenST(type->GetLoadSize());
		}
	}

	void GenST(wt_u32 loadsize)
	{
		switch(loadsize)
		{
		case sizeof(wt_u8):  mEngine.AddSingle(WT_TOMB_OP_STB); break;
		case sizeof(wt_u16): mEngine.AddSingle(WT_TOMB_OP_STH); break;
		default:             mEngine.MakeST(loadsize);  break;
		}
	}

	void GenLD(wtast::Type* type)
	{
		if(type->IsFunctionType())
		{
			mEngine.AddSingle(WT_TOMB_OP_LDFP);
		}
		else
		{
			GenLD(type->GetLoadSize());
		}
	}

	void GenLD(wt_u32 loadsize)
	{
		switch(loadsize)
		{
		case sizeof(wt_u8):  mEngine.AddSingle(WT_TOMB_OP_LDB); break;
		case sizeof(wt_u16): mEngine.AddSingle(WT_TOMB_OP_LDH); break;
		default:             mEngine.MakeLD(loadsize);  break;
		}
	}

	void GenAddImm(wt_u32 loadsize, wt_u8 imm)
	{
		switch(loadsize)
		{
		case sizeof(wt_u8):
			mEngine.AddSingle(WT_TOMB_OP_ADDB_I, imm);
			break;

		case sizeof(wt_u16):
			mEngine.AddSingle(WT_TOMB_OP_ADDH_I, imm);
			break;

		case sizeof(wt_u32):
			mEngine.AddSingle(WT_TOMB_OP_ADD_I, imm);
			break;

		default:
			mEngine.AddSingle(WT_TOMB_OP_PWZL_I, imm);
			mEngine.AddSingle(WT_TOMB_OP_ADDL);
			break;
		}
	}

	void GenSubImm(wt_u32 loadsize, wt_u8 imm)
	{
		switch(loadsize)
		{
		case sizeof(wt_u8):
			mEngine.AddSingle(WT_TOMB_OP_SUBB_I, imm);
			break;

		case sizeof(wt_u16):
			mEngine.AddSingle(WT_TOMB_OP_SUBH_I, imm);
			break;

		case sizeof(wt_u32):
			mEngine.AddSingle(WT_TOMB_OP_SUB_I, imm);
			break;

		default:
			mEngine.AddSingle(WT_TOMB_OP_PWZL_I, imm);
			mEngine.AddSingle(WT_TOMB_OP_SUBL);
			break;
		}
	}

	void AddPostCondFooter(wtast::PostCondition* postcond)
	{
		wt_u32 sum = 0;

		// summarize the total size of the return value

		for(std::size_t i=0,n=postcond->GetNumReturnValues();i!=n;++i)
		{
			wtast::VariableDeclaration* decl = postcond->GetReturnValue(i);

			sum += CodeGen::AlignSizeUp(decl->GetType()->GetLoadSize()) * decl->GetNumberOfFragments();
		}

		if(0<sum)
		{
			// duplicate the data on the data stack

			mEngine.MakeDUP(sum);

			// store the values (by reference)

			for(std::size_t i=0,n=postcond->GetNumReturnValues();i!=n;++i)
			{
				wtast::VariableDeclaration* decl = postcond->GetReturnValue(n-i-1);
				wtast::Type*                type = decl->GetType();

				for(std::size_t j=0,m=decl->GetNumberOfFragments();j!=m;++j)
				{
					wtast::VariableDeclFragment* frag = decl->GetFragment(m-j-1);

					const FrameStack::Variable* var = mFrameStack.FindVar(frag->GetUniqueId());

					if(var)
					{
						GenPushVarAddr(var);
						GenST(type);
					}
					else
					{
						mCodeGen.AddError(postcond, std::string("Internal error: return variable '") + frag->GetName() + "' not found.");
					}
				}
			}
		}

		// check the condition

		if(!mExternals)
		{
			mCodeGen.AddError(postcond, "Internal error: externals section missing");
			return;
		}

		std::string lab_after = mEngine.GenLabel();

		postcond->GetCondition()->Accept(this);
		mEngine.AddJTI(lab_after);
		mEngine.MakePBA(mExternals->GetExcAddr(mCodeGen.PostcondViolExc));
		mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
		mEngine.AddSingle(WT_TOMB_OP_ERAISE);
		mEngine.AddLabel(lab_after);
	}

	void MakeTDR()
	{
		if(mEngine.LastIsWith(WT_TOMB_OP_TPU))
		{
			mEngine.RemoveLastInstrument();
			mEngine.MakePOP(sizeof(wtr_objentry));
		}
		else
		{
			mEngine.AddSingle(WT_TOMB_OP_TDR);
		}
	}

	void VisitSwitchDataType(wtast::SwitchStatement* sw, wtast::DataType* dt)
	{
		const bool is_temporary = sw->GetCondition()->IsTemporary();
		wt_u32 offs = 0u;

		if(is_temporary)
		{
			EnterFrame("");
			mFrameStack.AddVariable(dt, "$_automatic_$", mCodeGen.GetNextUniqueId(), offs);
		}

		sw->GetCondition()->Accept(this);

		if(is_temporary)
		{
			mEngine.MakeDUP(CodeGen::AlignSizeUp(dt->GetLoadSize()));
			mEngine.MakePVT(offs);
			GenST(dt);
			// 'dt' is always object
			MakeTDR();
			mEngine.MakePVT(offs);
			mCodeGen.GenDestructorMkfp(dt);
			mEngine.AddSingle(WT_TOMB_OP_OSPU);
		}

		const std::string end_label = mEngine.GenLabel();

		mEngine.MakeDUP(sizeof(wt_vmptr));
		mEngine.MakeRECFR(0);
		mEngine.MakeLD(sizeof(wt_u32));

		for(std::size_t i=0,n=sw->GetNumberOfCases();i!=n;++i)
		{
			wtast::SwitchCase* sc = sw->GetCase(i);

			if(wtast::AstNode::DATATYPECASESTATEMENT==sc->GetKind())
			{
				wtast::DataTypeCaseStatement* dtcs = static_cast<wtast::DataTypeCaseStatement*>(sc);
				wtast::DataTypeCons*          cons = dtcs->GetCons();

				const std::string nextcons_label = mEngine.GenLabel();

				mEngine.MakeDUP(sizeof(wt_u32));
				mEngine.MakePushImm(cons->GetConsId());
				mEngine.AddSingle(WT_TOMB_OP_CMP);
				mEngine.AddJTL(nextcons_label);

				if(mFrameStack.EnterFrame(""))
				{
					mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));

					for(std::size_t i=0,n=cons->GetNumberOfTypes();i!=n;++i)
					{
						wtast::Type* type = cons->GetType(i);

						mEngine.MakeDUP(sizeof(wt_vmptr));
						mEngine.MakeRECFR(cons->GetOffset(i));
						GenLD(type);

						wtast::VariableDeclaration* vardecl = dtcs->GetVariable(i);
						if(1!=vardecl->GetNumberOfFragments())
						{
							mCodeGen.AddError(dtcs, "Variable declaration has multiple fragments.");
							continue;
						}

						wtast::VariableDeclFragment* frag = vardecl->GetFragment(0);

						bool   succeeded;
						wt_u32 var_offs = 0;

						if(type->IsObject())
						{
							succeeded = mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), frag->GetName(), frag->GetUniqueId(), var_offs);
						}
						else
						{
							succeeded = mFrameStack.AddVariable(frag, false, var_offs);
						}

						if(!succeeded)
						{
							mCodeGen.AddError(dtcs, std::string("Failed to add variable '") + frag->GetName() + "'.");
							continue;
						}

						mEngine.MakePVT(var_offs);
						GenST(type);
					}

					mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_vmptr));
					dtcs->GetBody()->Accept(this);

					mFrameStack.LeaveFrame();
				}
				else
				{
					mCodeGen.AddError(dtcs, "Failed to create temporary frame");
				}

				if(Instrument::IK_JMPL!=mEngine.GetLastInstrument().kind)
				{
					mEngine.AddJmpL(end_label);
				}

				mEngine.AddLabel(nextcons_label);
			}
			else
			if(wtast::AstNode::DEFAULTSTATEMENT==sc->GetKind())
			{
				mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32)+sizeof(wt_vmptr));
				sc->Accept(this);

				if(Instrument::IK_JMPL!=mEngine.GetLastInstrument().kind)
				{
					mEngine.AddJmpL(end_label);
				}
			}
			else
			{
				mCodeGen.AddError(sc, "Mixed switch cases are not supported");
			}
		}

		mEngine.AddLabel(end_label);
		if(is_temporary)
		{
			LeaveFrame();
		}
	}

	void VisitSwitchEnumType(wtast::SwitchStatement* sw, wtast::EnumType* et)
	{
		sw->GetCondition()->Accept(this);

		const std::string end_label = mEngine.GenLabel();

		for(std::size_t i=0,n=sw->GetNumberOfCases();i!=n;++i)
		{
			wtast::SwitchCase* sc = sw->GetCase(i);

			if(wtast::AstNode::ENUMCASESTATEMENT==sc->GetKind())
			{
				wtast::EnumCaseStatement* ecs        = static_cast<wtast::EnumCaseStatement*>(sc);
				wtast::Enumerator*        enumerator = ecs->GetEnumerator();

				const std::string nextenumerator_label = mEngine.GenLabel();

				mEngine.MakeDUP(sizeof(wt_u32));
				mEngine.MakePushImm(enumerator->GetOrdinal());
				mEngine.AddSingle(WT_TOMB_OP_CMP);
				mEngine.AddJTL(nextenumerator_label);

				mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));
				ecs->GetBody()->Accept(this);

				if(Instrument::IK_JMPL!=mEngine.GetLastInstrument().kind)
				{
					mEngine.AddJmpL(end_label);
				}

				mEngine.AddLabel(nextenumerator_label);
			}
			else
			if(wtast::AstNode::DEFAULTSTATEMENT==sc->GetKind())
			{
				mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));
				sc->Accept(this);

				if(Instrument::IK_JMPL!=mEngine.GetLastInstrument().kind)
				{
					mEngine.AddJmpL(end_label);
				}
			}
			else
			{
				mCodeGen.AddError(sc, "Mixed switch cases are not supported");
			}
		}

		mEngine.AddLabel(end_label);
	}

public:
	GenVisitor(CodeGen & cg, InstrumentEngine & eng, DataSection* datasec, ExternalsSection* extsec, VisitMode visitmode)
		: wtast::AstVisitor()
		, mCodeGen(cg)
		, mEngine(eng)
		, mDataSection(datasec)
		, mExternals(extsec)
		, mVisitMode(visitmode)
		, mInGlobalBlock(false)
		, mFuncDef(0)
		, mPostCond(0)
	{
		mCodeGen.FillGlobalVariables(mFrameStack);
	}

	bool Visit(wtast::MainBlock* mb)
	{
		EnterFrame(CodeGen::MainFuncName);
		return true;
	}

	void EndVisit(wtast::MainBlock*)
	{
		LeaveFrame();
	}

	bool Visit(wtast::ConstructorBlock* cb)
	{
		EnterFrame(CodeGen::CtorFuncName);
		return true;
	}

	void EndVisit(wtast::ConstructorBlock*)
	{
		LeaveFrame();
	}

	bool Visit(wtast::DestructorBlock* db)
	{
		EnterFrame(CodeGen::DtorFuncName);
		return true;
	}

	void EndVisit(wtast::DestructorBlock*)
	{
		LeaveFrame();
	}

	bool Visit(wtast::FunctionDefinition* funcdef)
	{
		if(AddAssert(mCodeGen.mProblemHandler, 0==mFuncDef))
		{
			return false;
		}

		mFuncDef = funcdef;

		wtast::FunctionSignature* sig = funcdef->GetFunctionSignature();

		EnterFrame(sig->GetCanonicalName(false, true));

		for(std::size_t i=0,n=sig->GetArgc();i!=n;++i)
		{
			mFrameStack.AddVariable(sig->GetArg(i));
		}

		for(std::size_t i=0,n=sig->GetArgc();i!=n;++i)
		{
			wtast::FunctionArgument* arg = sig->GetArg(n-i-1);
			const FrameStack::Variable* var = mFrameStack.FindVar(arg->GetUniqueId());

			mEngine.MakePVT(var->offset);
			GenST(arg->GetType());
		}

		if(funcdef->GetPreCondition())
		{
			funcdef->GetPreCondition()->Accept(this);
		}

		if(funcdef->GetPostCondition())
		{
			funcdef->GetPostCondition()->Accept(this);
		}

		EnterFrame(sig->GetName());

		funcdef->GetBody()->Accept(this);

		LeaveFrame();
		LeaveFrame();

		mFuncDef = 0;

		return false;
	}

/*
	void EndVisit(wtast::FunctionDefinition*)
	{
	}
*/

	bool Visit(wtast::Compound* c)
	{
		EnterFrame(c->GetName());
		return true;
	}

	void EndVisit(wtast::Compound*)
	{
		LeaveFrame();
	}

/*
	bool Visit(wtast::EHandlers* ehs)
	{
		return true;
	}
*/

	bool Visit(wtast::EHandler* eh)
	{
		std::string lab_nexteh;

		const std::string excname = eh->GetName();
		bool need_cmp = "..."!=excname;

		if(need_cmp)
		{
			lab_nexteh = mEngine.GenLabel();

			mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_exception));

			const wt_u32 addr = mExternals->GetExcAddr(excname);

			if(ExternalsSection::BadAddr!=addr)
			{
				mEngine.MakePBA(addr);
			}
			else
			{
				std::ostringstream ss;
				ss << "The exception '";
				ss << excname;
				ss << "' can not be resolved.";
				mCodeGen.AddError(eh, ss.str());
			}

			mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
			mEngine.AddSingle(WT_TOMB_OP_CMP);
			mEngine.AddSingle(WT_TOMB_OP_QEQ);
			mEngine.AddJFL(lab_nexteh);
		}

		mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_exception));
		eh->GetBody()->Accept(this);
		mEngine.AddJmpL(mLabelAftereh);

		if(need_cmp)
		{
			mEngine.AddLabel(lab_nexteh);
		}

		return false;
	}

	bool Visit(wtast::ExceptionCompound* ec)
	{
		bool have_eh_or_finally = HaveEHandlers(ec) || HaveFinally(ec);

		if(have_eh_or_finally)
		{
			mLabelExhan = mEngine.GenLabel();
			mLabelAftereh = mEngine.GenLabel();
			mEngine.AddExhan(mLabelExhan);
		}

		if(HaveFinally(ec))
		{
			mLabelFinally = mEngine.GenLabel();
		}

		mLabelCleanup = mEngine.GenLabel();

		for(std::size_t i=0,n=ec->GetNumberOfStatements();i!=n;++i)
		{
			ec->GetStatement(i)->Accept(this);
		}

		if(have_eh_or_finally)
		{
			mEngine.AddJmpL(mLabelCleanup);
			mEngine.AddLabel(mLabelExhan);

			mEngine.AddSingle(WT_TOMB_OP_EHCL);
			mEngine.AddSingle(WT_TOMB_OP_TCL);

			if(HaveEHandlers(ec))
			{
				ec->GetEHandlers()->Accept(this);
			}

			mEngine.AddSingle(WT_TOMB_OP_ERAISE);

			if(HaveFinally(ec))
			{
				mEngine.AddLabel(mLabelFinally);
				ec->GetFinally()->Accept(this);
				mEngine.AddSingle(WT_TOMB_OP_JA_S);
			}

			mEngine.AddLabel(mLabelAftereh);
		}

		mEngine.AddLabel(mLabelCleanup);
		CallFinallyIfNeeded();

		return false;
	}

	bool Visit(wtast::VariableDeclaration* vardecl)
	{
		if(mInGlobalBlock && !mCodeGen.IsReferredGlobal(vardecl))
		{
			return false;
		}

		return true;
	}

	void EndVisit(wtast::VariableDeclaration* vardecl)
	{
		if(mInGlobalBlock && VISIT_NORMAL==mVisitMode)
		{
			return;
		}

		wtast::Type* type = vardecl->GetType();
		const bool type_is_object = type->IsObject();
		wtast::Expression* const init = vardecl->GetInitialValue();

		if(!type_is_object && !type->IsDefaultConstructible() && !init)
		{
			std::ostringstream ss;
			ss << "The type '";
			ss << type->GetName();
			ss << "' is not default constructible.";
			mCodeGen.AddError(vardecl, ss.str());
			return;
		}

		for(std::size_t i=0,n=vardecl->GetNumberOfFragments();i!=n;++i)
		{
			const FrameStack::Variable* var = mFrameStack.FindVar(vardecl->GetFragment(n-i-1)->GetUniqueId());

			if(type_is_object)
			{
				if(0==init)
				{
					mCodeGen.GenConstructorCall(type);
				}
				else
				if(!init->GetExpression(n-i-1)->IsTemporary())
				{
					mCodeGen.GenCopyCall(type);
				}
			}
			else
			if(0==init)
			{
				switch(type->GetLoadSize())
				{
				case 4: mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);  break;
				case 8: mEngine.AddSingle(WT_TOMB_OP_PWZL_I, 0); break;
				default:
					mCodeGen.AddError(vardecl, std::string("Type '") +
						type->GetName() + "' is not default contructible.");
					break;
				}
			}

			GenPushVarAddr(var);
			GenST(type);

			if(type_is_object)
			{
				MakeTDR();

				if(VISIT_NORMAL==mVisitMode)
				{
					GenPushVarAddr(var);
					mCodeGen.GenDestructorMkfp(type);
					mEngine.AddSingle(WT_TOMB_OP_OSPU);
				}
			}
		}
	}

	bool Visit(wtast::VariableDeclFragment* frag)
	{
		if(VISIT_NORMAL==mVisitMode)
		{
			mFrameStack.AddVariable(frag, false);
		}
		else
		if(VISIT_GLOBALCTOR==mVisitMode)
		{
			mFrameStack.AddVariable(frag, true);
			mCodeGen.AddGlobalVariable(frag, mFrameStack.GetGlobalDvtSize());
		}
		return false;
	}

	bool Visit(wtast::GlobalBlock*)
	{
		if(VISIT_GLOBALCTOR==mVisitMode)
		{
			EnterFrame(CodeGen::GlobalCtorFuncName);
		}

		mInGlobalBlock = true;
		return true;
	}

	void EndVisit(wtast::GlobalBlock*)
	{
		if(VISIT_GLOBALCTOR==mVisitMode)
		{
			LeaveFrame();
		}

		mInGlobalBlock = false;
	}

	bool Visit(wtast::VariableAccess* varacc)
	{
		const FrameStack::Variable* var = mFrameStack.FindVar(varacc->GetUniqueId());

		if(!var)
		{
			mCodeGen.AddError(varacc, std::string("The variable '") + varacc->GetName() + "' not found.");
			return false;
		}

		GenPushVarAddr(var);

		wtast::TypeInfo* typeinfo = varacc->GetType();

		if(typeinfo->IsSimultaneous())
		{
			mCodeGen.AddError(varacc,
				"Internal error: type of a variable can not be simultaneous.");
			return false;
		}

		GenLD(static_cast<wtast::Type*>(typeinfo));

		return false;
	}

/*
	bool Visit(wtast::ElementAccess*)
	{
		return true;
	}
*/

	void EndVisit(wtast::ElementAccess* elemaccess)
	{
		wtast::TypeInfo* ltypeinfo = elemaccess->GetLValue()->GetType();
		wtast::Type*     ltype     = ltypeinfo->GetType(0);

		if(ltypeinfo->IsSimultaneous() || (!ltype->IsSequence() && WTT_STRING!=(ltype->GetTypeId() & ~(WTTF_MUTABLE/*|WTTF_REF*/))))
		{
			mCodeGen.AddError(elemaccess, std::string("Can not access element of type '") + ltypeinfo->GetName() + "'.");
			return;
		}

		wtast::TypeInfo* itypeinfo = elemaccess->GetIndexing()->GetType();
		wtast::Type*     itype     = itypeinfo->GetType(0);

		if(itypeinfo->IsSimultaneous())
		{
			mCodeGen.AddError(elemaccess, std::string("Simultaneous expression is not allowed in indexing: '") + itypeinfo->GetName() + "'.");
			return;
		}

		bool is_signed = false;
		bool is_large  = false;

		switch(itype->GetTypeId())
		{
		case WTT_U8:
		case WTT_U16:
		case WTT_U32:
			// no-op, this is good as it is
			break;

		// Don't waste instructions on this, so let the microprogram
		// handle everything, i.e. negative numbers, and too large numbers.
		// The final result will be exactly the same if the user wrote this:
		// :: expression [ toTYPE( expression ) ]

		case WTT_I8:
			is_signed = true;
			mEngine.AddSingle(WT_TOMB_OP_SGNXD, 8);
			mEngine.AddSingle(WT_TOMB_OP_EXEC_I, WTMP_CI32U32);
			break;

		case WTT_I16:
			is_signed = true;
			mEngine.AddSingle(WT_TOMB_OP_SGNXD, 16);
			mEngine.AddSingle(WT_TOMB_OP_EXEC_I, WTMP_CI32U32);
			break;

		case WTT_I32:
			is_signed = true;
			mEngine.AddSingle(WT_TOMB_OP_EXEC_I, WTMP_CI32U32);
			break;

		case WTT_U64:
			is_large = true;
			mEngine.AddSingle(WT_TOMB_OP_EXEC_I, WTMP_CU64U32);
			break;

		case WTT_I64:
			is_signed = true;
			is_large  = true;
			mEngine.AddSingle(WT_TOMB_OP_EXEC_I, WTMP_CI64U32);
			break;

		default:
			mCodeGen.AddError(elemaccess, std::string("Invalid indexing expression type: '") + itypeinfo->GetName() + "'.");
			return;
		}

		if(is_signed && !mCodeGen.GetOptions()->SignedIndex())
		{
			mCodeGen.AddError(elemaccess, std::string("Indexing with signed integer type ('") + itypeinfo->GetName() + "') is not allowed.");
		}
		if(is_large && !mCodeGen.GetOptions()->LargeIndex())
		{
			mCodeGen.AddError(elemaccess, std::string("Indexing with 64-bit integer type ('") + itypeinfo->GetName() + "') is not allowed.");
		}

		if(ltype->IsSequence())
		{
			mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
			GenLD(static_cast<wtast::SeqType*>(ltype)->GetElementType());
		}
		else
		{
			mEngine.AddSingle(WT_TOMB_OP_STRCR_S);
			mEngine.AddSingle(WT_TOMB_OP_LDB);
		}
	}

/*
	bool Visit(wtast::MemberAccess* memberaccess)
	{
		return true;
	}
*/

	void EndVisit(wtast::MemberAccess* member)
	{
		if(member->GetType()->IsSimultaneous())
		{
			mCodeGen.AddError(member, std::string("Type of record field '") +
				member->GetFieldName() + "' is simultaneous type.");
			return;
		}

		mEngine.MakeRECFR(member->GetOffset());
		GenLD(member->GetType()->GetType(0));
	}

	bool Visit(wtast::BasicType* basictype)
	{
		return false;
	}

	bool Visit(wtast::SeqType* seqtype)
	{
		return true;
	}

	void EndVisit(wtast::SeqType*)
	{
	}

	bool Visit(wtast::SimultaneousType* simtype)
	{
		return true;
	}

	void EndVisit(wtast::SimultaneousType*)
	{
	}

	bool Visit(wtast::RecordField* recfield)
	{
		return true;
	}

	void EndVisit(wtast::RecordField*)
	{
	}

	bool Visit(wtast::RecordType* rec)
	{
		return true;
	}

	void EndVisit(wtast::RecordType*)
	{
	}

	bool Visit(wtast::RecordDeclaration* recdecl)
	{
		return true;
	}

	void EndVisit(wtast::RecordDeclaration*)
	{
	}

	bool Visit(wtast::FunctionType*)
	{
		return true;
	}

	void EndVisit(wtast::FunctionType*)
	{
	}

	bool Visit(wtast::SimultaneousExpression* simexpr)
	{
		return true;
	}

	void EndVisit(wtast::SimultaneousExpression*)
	{
	}

	bool Visit(wtast::StringLiteral* lit)
	{
		mEngine.MakePBA(mDataSection->Find(lit->GetLiteral()));
		mEngine.AddSingle(WT_TOMB_OP_STRN);
		mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
		mCodeGen.GenDestructorMkfp(static_cast<wtast::Type*>(lit->GetType()));
		mEngine.AddSingle(WT_TOMB_OP_TPU);
		return false;
	}

	bool Visit(wtast::BooleanLiteral* lit)
	{
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, lit->GetLiteral());
		return false;
	}

	bool Visit(wtast::CharacterLiteral* lit)
	{
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, wt_u8(lit->GetLiteral()));
		return false;
	}

	bool Visit(wtast::NumberLiteral* lit)
	{
		if(lit->Is64Bit())
		{
			if(mCodeGen.GetOptions()->Inline64())
			{
				mEngine.MakePushImm(lit->GetLiteralU64());
			}
			else
			{
				mEngine.MakePBA(mDataSection->Find(lit->GetLiteralU64()));
				mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_u64));
			}
		}
		else
		{
			if(mCodeGen.GetOptions()->Inline32())
			{
				mEngine.MakePushImm(lit->GetLiteralU32());
			}
			else
			{
				mEngine.MakePBA(mDataSection->Find(lit->GetLiteralU32()));
				mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_u32));
			}
		}

		return false;
	}

	bool Visit(wtast::EnumLiteral* lit)
	{
		mEngine.MakePushImm(lit->GetEnumerator()->GetOrdinal());
		return false;
	}

	bool Visit(wtast::MakeFunctionPointer* mkfp)
	{
		wtast::FunctionSignature* funcsig = mkfp->GetFunctionSignature();
		const std::string canonical_name = funcsig->GetCanonicalName(false, true);

		if(funcsig->IsBuiltin())
		{
			mEngine.AddBuiltinInstantiation(funcsig->GetUniqueId(), canonical_name);
			mEngine.AddMkfp(canonical_name);
		}
		else
		if(mExternals && mExternals->HasFunc(canonical_name))
		{
			const wt_u32 addr = mExternals->GetFuncAddr(canonical_name);

			if(ExternalsSection::BadAddr!=addr)
			{
				mEngine.MakePBA(addr);
				mEngine.AddSingle(WT_TOMB_OP_LDFP);
			}
			else
			{
				std::ostringstream ss;
				ss << "The function '";
				ss << canonical_name;
				ss << "' can not be resolved.";
				mCodeGen.AddError(mkfp, ss.str());
			}
		}
		else
		{
			if(funcsig->IsGenerated())
			{
				mCodeGen.RegisterGeneratedFunction(funcsig);
			}

			mEngine.AddMkfp(canonical_name);
		}

		return false;
	}

	bool Visit(wtast::FunctionPointerCall* fpc)
	{
		if(fpc->GetArgumentExpression())
		{
			fpc->GetArgumentExpression()->Accept(this);
		}

		fpc->GetFunctionPointerAccess()->Accept(this);
		mEngine.AddSingle(WT_TOMB_OP_CFP);
		return false;
	}

	bool Visit(wtast::Assignment* assign)
	{
		if(!assign->GetExpression())
		{
			mCodeGen.AddError(assign, "Assignment does not have expression.");
			return false;
		}

		wtast::Expression* expr = assign->GetExpression();

		std::deque<wtast::LValue*> lvals;
		for(std::size_t i=0,n=assign->GetNumberOfLValues();i!=n;++i)
		{
			lvals.push_back(assign->GetLValue(i));
		}
		std::deque<wtast::LValue*> lvals2 = lvals;

		for(std::size_t ii=0,nn=expr->GetNumberOfExpressions();ii!=nn;++ii)
		{
			wtast::Expression* rhs = expr->GetExpression(ii);

			rhs->Accept(this);

			for(std::size_t i=0,n=rhs->GetType()->GetNumberOfTypes();i!=n;++i)
			{
				wtast::LValue* lvalue = lvals2.front();
				lvals2.pop_front();

				wtast::Type* rtype = rhs->GetType()->GetType(i);

				if(!lvalue->GetType())
				{
					mCodeGen.AddError(lvalue, "TODO: assignment lvalue->GetType() is null");
					return false;
				}

				if(lvalue->GetType()->IsSimultaneous())
				{
					mCodeGen.AddError(lvalue, "TODO: assignment lvalue->GetType() is simultaneous");
					return false;
				}

				wtast::Type* ltype = static_cast<wtast::Type*>(lvalue->GetType());

				if(ltype->GetTypeId()!=rtype->GetTypeId())
				{
					mCodeGen.AddError(lvalue, "TODO: assignment lvalue->GetType() and expr->GetType() mismatch");
					return false;
				}

				if(ltype->IsObject())
				{
					if(!rhs->IsTemporary())
					{
						mCodeGen.GenCopyCall(ltype);
					}

					MakeTDR();
				}
			}
		}

		if(lvals.size()!=expr->GetType()->GetNumberOfTypes())
		{
			mCodeGen.AddError(assign, "Assignment arity does not match.");
			return false;
		}

		for(std::size_t ii=0,nn=expr->GetNumberOfExpressions();ii!=nn;++ii)
		{
			wtast::Expression* rhs = expr->GetExpression(nn-ii-1);

			for(std::size_t i=0,n=rhs->GetType()->GetNumberOfTypes();i!=n;++i)
			{
				wtast::LValue* lvalue = lvals.back();
				lvals.pop_back();

// unused?
//				wtast::Type* rtype = rhs->GetType()->GetType(n-i-1);
				wtast::Type* ltype = static_cast<wtast::Type*>(lvalue->GetType());

				lvalue->Accept(this);

				if(ltype->IsObject())
				{
					Instrument last = mEngine.GetLastInstrument();
					mEngine.RemoveLastInstrument();

					if(mEngine.LastIsWith(WT_TOMB_OP_LD_S))
					{
						mEngine.MakePDSP(sizeof(wt_u32));
						mEngine.AddSingle(WT_TOMB_OP_IDRD_I);
					}
					else
					{
						mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
					}

					mEngine.Add(last);
					mCodeGen.GenDestructorCall(ltype);
					mEngine.Add(last);
				}

				if(!mEngine.ReplaceLast(WT_TOMB_OP_LD_I, WT_TOMB_OP_ST_I) &&
					!mEngine.ReplaceLast(WT_TOMB_OP_LD_S, WT_TOMB_OP_ST_S) &&
					!mEngine.ReplaceLast(WT_TOMB_OP_LD_L, WT_TOMB_OP_ST_L) &&
					!mEngine.ReplaceLast(WT_TOMB_OP_LDB, WT_TOMB_OP_STB) &&
					!mEngine.ReplaceLast(WT_TOMB_OP_LDFP, WT_TOMB_OP_STFP))
				{
					mCodeGen.AddError(lvalue, "TODO: assignment ReplaceLast() failed");
				}
			}
		}

		mEngine.AddSingle(WT_TOMB_OP_TCL);
		return false;
	}

/*
	bool Visit(wtast::FunctionSignature* funcsig)
	{
		return true;
	}

	void EndVisit(wtast::FunctionSignature*)
	{
	}

	bool Visit(wtast::FunctionArgument* funcarg)
	{
		return true;
	}

	void EndVisit(wtast::FunctionArgument*)
	{
	}

	bool Visit(wtast::FunctionCall*)
	{
		return true;
	}
*/

	bool Visit(wtast::FunctionDeclaration* funcdecl)
	{
		return true;
	}

	void EndVisit(wtast::FunctionDeclaration*)
	{
	}

	void EndVisit(wtast::FunctionCall* funccall)
	{
		wtast::FunctionSignature* sig = funccall->GetFunctionSignature();
		wtast::Expression*        arg = funccall->GetExpression();

		std::size_t num_tmp_arg = 0;

		if(arg)
		{
			for(std::size_t i=0,n=arg->GetNumberOfExpressions();i<n;++i)
			{
				if(arg->GetExpression(i)->GetType()->HasObject() && arg->GetExpression(i)->IsTemporary())
				{
					++num_tmp_arg;
				}
			}
		}

		if(sig->IsBuiltin())
		{
			mEngine.MakeEXEC(sig);
		}
		else
		{
			const std::string canonical_name = sig->GetCanonicalName(false, true);

			if(mExternals && mExternals->HasFunc(canonical_name))
			{
				const wt_u32 addr = mExternals->GetFuncAddr(canonical_name);

				if(ExternalsSection::BadAddr!=addr)
				{
					mEngine.MakeDLE(addr);
				}
				else
				{
					std::ostringstream ss;
					ss << "The function '";
					ss << canonical_name;
					ss << "' can not be resolved.";
					mCodeGen.AddError(funccall, ss.str());
				}
			}
			else
			{
				if(sig->IsGenerated())
				{
					mCodeGen.RegisterGeneratedFunction(sig);
				}

				mEngine.MakeENTER(sig);
			}
		}

		std::size_t num_obj_ret = 0;

		for(std::size_t i=0,n=sig->GetRetc();i<n;++i)
		{
			if(sig->GetRet(i)->IsObject())
			{
				mEngine.AddSingle(WT_TOMB_OP_TPO);
				++num_obj_ret;
			}
		}

		for(std::size_t i=0;i<num_tmp_arg;++i)
		{
			mEngine.AddSingle(WT_TOMB_OP_TTD);
		}

		for(std::size_t i=0;i<num_obj_ret;++i)
		{
			if(mEngine.LastIsWith(WT_TOMB_OP_TPO))
			{
				mEngine.RemoveLastInstrument();
			}
			else
			{
				mEngine.AddSingle(WT_TOMB_OP_TPU);
			}
		}
	}

/*
	bool Visit(wtast::ProcedureCall*)
	{
		return true;
	}

	void EndVisit(wtast::ProcedureCall* pcall)
	{
	}
*/

	bool Visit(wtast::RaiseStatement* raisestmt)
	{
		if(!raisestmt->IsNear())
		{
			mEngine.AddSingle(WT_TOMB_OP_EHCL);
		}

		CallFinallyIfNeeded();

		wt_u32 exc_addr;

		if(mExternals &&
			ExternalsSection::BadAddr!=(exc_addr=
				mExternals->GetExcAddr(raisestmt->GetException())))
		{
			mEngine.MakePBA(exc_addr);
		}
		else
		{
			std::ostringstream ss;
			ss << "The exception '";
			ss << raisestmt->GetException();
			ss << "' can not be resolved.";
			mCodeGen.AddError(raisestmt, ss.str());
		}

		mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
		mEngine.AddSingle(WT_TOMB_OP_ERAISE);

		return false;
	}

	bool Visit(wtast::UnresolvedFunctionName* ufn)
	{
		mCodeGen.AddError(ufn, "Unresolved function name found.");
		return false;
	}

	bool Visit(wtast::PreCondition* precond)
	{
		if(mCodeGen.GetOptions()->EnablePrecond())
		{
			if(AddAssert(mCodeGen.mProblemHandler, 0!=mExternals) ||
				AddAssert(mCodeGen.mProblemHandler,
					mExternals->HasException(mCodeGen.PrecondViolExc)))
			{
				return false;
			}

			std::string lab_after = mEngine.GenLabel();

			precond->GetCondition()->Accept(this);
			mEngine.AddJTI(lab_after);
			mEngine.MakePBA(mExternals->GetExcAddr(mCodeGen.PrecondViolExc));
			mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
			mEngine.AddSingle(WT_TOMB_OP_ERAISE);
			mEngine.AddLabel(lab_after);
		}

		return false;
	}

/*
	void EndVisit(wtast::PreCondition*)
	{
	}
*/

	bool Visit(wtast::PostCondition* postcond)
	{
		if(mCodeGen.GetOptions()->EnablePostcond())
		{
			if(AddAssert(mCodeGen.mProblemHandler, 0!=mExternals) ||
				AddAssert(mCodeGen.mProblemHandler,
					mExternals->HasException(mCodeGen.PostcondViolExc)) ||
				AddAssert(mCodeGen.mProblemHandler, 0==mPostCond))
			{
				return false;
			}

			for(std::size_t i=0,n=postcond->GetNumReturnValues();i!=n;++i)
			{
				wtast::VariableDeclaration* decl = postcond->GetReturnValue(i);
				wtast::Type*                type = decl->GetType();

				for(std::size_t j=0,m=decl->GetNumberOfFragments();j!=m;++j)
				{
					wtast::VariableDeclFragment* frag = decl->GetFragment(j);
					bool                         succ = false;

					if(type->IsObject())
					{
						succ = mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), frag->GetName(), frag->GetUniqueId());
					}
					else
					{
						succ = mFrameStack.AddVariable(frag, false);
					}

					if(!succ)
					{
						mCodeGen.AddError(postcond, "Internal error: failed to add return variable.");
					}
				}
			}

			for(std::size_t i=0,n=postcond->GetNumOldValues();i!=n;++i)
			{
				wtast::VariableDeclaration* decl = postcond->GetOldValue(i);
				wtast::Type*                type = decl->GetType();

				for(std::size_t j=0,m=decl->GetNumberOfFragments();j!=m;++j)
				{
					wtast::VariableDeclFragment* frag = decl->GetFragment(j);
					wt_u32                       offs = 0;
					std::string                  name = frag->GetName().substr(4);
					const FrameStack::Variable*  var  = FindVarInFormals(name);

					if(!mFrameStack.AddVariable(frag, false, offs))
					{
						mCodeGen.AddError(postcond, "Internal error: failed to add old variable.");
						continue;
					}

					if(var)
					{
						GenPushVarAddr(var);
						mEngine.MakeLD(sizeof(wt_vmptr));

						if(type->IsObject())
						{
							mCodeGen.GenCopyCall(type);
							mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
							mEngine.MakePVT(offs);
							mCodeGen.GenDestructorMkfp(type);
							MakeTDR();
							mEngine.AddSingle(WT_TOMB_OP_OSPU);
						}

						mEngine.MakePVT(offs);
						mEngine.MakeST(sizeof(wt_vmptr));
					}
					else
					{
						mCodeGen.AddError(postcond, std::string("Internal error: formal variable '") + name + "' not found.");
					}
				}
			}

			mPostCond = postcond;
		}

		return false;
	}

/*
	void EndVisit(wtast::PostCondition*)
	{
	}
*/

	bool Visit(wtast::AndThen* andthen)
	{
		std::string lab_then = mEngine.GenLabel();
		std::string lab_end  = mEngine.GenLabel();

		andthen->GetLeftHandSide()->Accept(this);
		mEngine.AddJTL(lab_then);
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, WT_FALSE);
		mEngine.AddJmpL(lab_end);
		mEngine.AddLabel(lab_then);
		andthen->GetRightHandSide()->Accept(this);
		mEngine.AddLabel(lab_end);

		return false;
	}

	bool Visit(wtast::OrElse* orelse)
	{
		std::string lab_then = mEngine.GenLabel();
		std::string lab_end  = mEngine.GenLabel();

		orelse->GetLeftHandSide()->Accept(this);
		mEngine.AddJFL(lab_then);
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, WT_TRUE);
		mEngine.AddJmpL(lab_end);
		mEngine.AddLabel(lab_then);
		orelse->GetRightHandSide()->Accept(this);
		mEngine.AddLabel(lab_end);

		return false;
	}

	bool Visit(wtast::ReturnStatement* retstmt)
	{
		mEngine.AddSingle(WT_TOMB_OP_TCL);

		if(retstmt->GetExpression())
		{
			wtast::Expression* expr = retstmt->GetExpression();
			expr->Accept(this);

			wtast::TypeInfo* typeinfo = expr->GetType();

			if(typeinfo && typeinfo->HasObject())
			{
				if(expr->IsSimultaneous())
				{
					wtast::SimultaneousExpression* simexpr = static_cast<wtast::SimultaneousExpression*>(expr);

					wt_u32 offs = 0;

					for(std::size_t i=0,n=simexpr->GetNumberOfExpressions();i!=n;++i)
					{
						wtast::Expression* subexpr = simexpr->GetExpression(n-i-1);

						CopyReturnValueIfNeeded(subexpr, offs);

						offs += static_cast<wtast::Type*>(subexpr->GetType())->GetLoadSize();
					}
				}
				else
				{
					CopyReturnValueIfNeeded(expr, 0);
				}
			}
		}

		GenCSOSDestructorCode();
		mEngine.AddJmpL(mLabelCleanup);
		return false;
	}

	bool Visit(wtast::WhileStatement* stmt)
	{
		std::string lab_begin = mEngine.GenLabel();
		std::string lab_end   = mEngine.GenLabel();

		mEngine.AddLabel(lab_begin);
		stmt->GetCondition()->Accept(this);
		mEngine.AddJFL(lab_end);
		stmt->GetBody()->Accept(this);
		mEngine.AddJmpL(lab_begin);
		mEngine.AddLabel(lab_end);

		return false;
	}

	bool Visit(wtast::DoWhileStatement* stmt)
	{
		std::string lab_begin = mEngine.GenLabel();

		mEngine.AddLabel(lab_begin);
		stmt->GetBody()->Accept(this);
		stmt->GetCondition()->Accept(this);
		mEngine.AddJTL(lab_begin);

		return false;
	}

	bool Visit(wtast::ForStatement* stmt)
	{
		std::string lab_begin = mEngine.GenLabel();
		std::string lab_end   = mEngine.GenLabel();

		if(stmt->GetInit())
		{
			stmt->GetInit()->Accept(this);
		}

		mEngine.AddLabel(lab_begin);

		if(stmt->GetCondition())
		{
			stmt->GetCondition()->Accept(this);
			mEngine.AddJFL(lab_end);
		}

		stmt->GetBody()->Accept(this);

		if(stmt->GetStep())
		{
			stmt->GetStep()->Accept(this);
		}

		mEngine.AddJmpL(lab_begin);
		mEngine.AddLabel(lab_end);

		return false;
	}

	bool Visit(wtast::ForeachStatement* foreach)
	{
		if(1!=foreach->GetVariable()->GetNumberOfFragments())
		{
			mCodeGen.AddError(foreach, "Only one variable declaration fragment is allowed.");
			return false;
		}

		EnterFrame("");

		if(foreach->IsRangeBased())
		{
			wt_u32 offs_i;
			wt_u32 offs_n;
			wt_u32 loadsize = foreach->GetVariable()->GetType()->GetLoadSize();
			wt_u8  cmp;

			switch(foreach->GetVariable()->GetType()->GetTypeId())
			{
			case WTT_CHAR: cmp = WT_TOMB_OP_CMP;   break;
			case WTT_I8:   cmp = WT_TOMB_OP_SCMPB; break;
			case WTT_U8:   cmp = WT_TOMB_OP_CMP;   break;
			case WTT_I16:  cmp = WT_TOMB_OP_SCMPH; break;
			case WTT_U16:  cmp = WT_TOMB_OP_CMP;   break;
			case WTT_I32:  cmp = WT_TOMB_OP_SCMP;  break;
			case WTT_U32:  cmp = WT_TOMB_OP_CMP;   break;
			case WTT_I64:  cmp = WT_TOMB_OP_SCMPL; break;
			case WTT_U64:  cmp = WT_TOMB_OP_CMPL;  break;
			default:
				mCodeGen.AddError(foreach, "Type is not supported.");
				LeaveFrame();
				return false;
			}

			mFrameStack.AddVariable(foreach->GetVariable()->GetFragment(0), false, offs_i);
			mFrameStack.AddVariable(foreach->GetVariable()->GetType(), "end", mCodeGen.GetNextUniqueId(), offs_n);

			std::string lab_loop = mEngine.GenLabel();
			std::string lab_step = mEngine.GenLabel();
			std::string lab_end  = mEngine.GenLabel();

			if(foreach->IsReverse())
			{
				foreach->GetFrom()->Accept(this);
				foreach->GetTo()->Accept(this);
			}
			else
			{
				foreach->GetTo()->Accept(this);
				foreach->GetFrom()->Accept(this);
			}

			mEngine.MakePVT(offs_i);
			GenST(loadsize);
			mEngine.MakePVT(offs_n);
			GenST(loadsize);

			if(foreach->IsReverse())
			{
				mEngine.MakePVT(offs_n);
				GenLD(loadsize);
				mEngine.MakePVT(offs_i);
				GenLD(loadsize);
			}
			else
			{
				mEngine.MakePVT(offs_i);
				GenLD(loadsize);
				mEngine.MakePVT(offs_n);
				GenLD(loadsize);
			}
			mEngine.AddSingle(cmp);
			if(foreach->IsInclusive())
			{
				mEngine.AddSingle(WT_TOMB_OP_QLE);
			}
			else
			{
				mEngine.AddSingle(WT_TOMB_OP_QLT);
			}
			mEngine.AddJFL(lab_end);

			if(foreach->IsReverse() && !foreach->IsInclusive())
			{
				mEngine.MakePVT(offs_i);
				GenLD(loadsize);
				GenSubImm(loadsize, 1);
				mEngine.MakePVT(offs_i);
				GenST(loadsize);
			}

			mEngine.AddLabel(lab_loop);

			if(foreach->IsReverse())
			{
				mEngine.MakePVT(offs_n);
				GenLD(loadsize);
				mEngine.MakePVT(offs_i);
				GenLD(loadsize);
			}
			else
			{
				mEngine.MakePVT(offs_i);
				GenLD(loadsize);
				mEngine.MakePVT(offs_n);
				GenLD(loadsize);
			}
			mEngine.AddSingle(cmp);
			if(!foreach->IsInclusive() && !foreach->IsReverse())
			{
				mEngine.AddSingle(WT_TOMB_OP_QLT);
			}
			else
			{
				mEngine.AddSingle(WT_TOMB_OP_QLE);
			}
			mEngine.AddJFL(lab_end);

			if(foreach->GetWhile())
			{
				foreach->GetWhile()->Accept(this);
				mEngine.AddJFL(lab_end);
			}

			if(foreach->GetWhich())
			{
				foreach->GetWhich()->Accept(this);
				mEngine.AddJFL(lab_step);
			}

			foreach->GetBody()->Accept(this);

			mEngine.AddLabel(lab_step);

			mEngine.MakePVT(offs_i);
			GenLD(loadsize);
			mEngine.MakePVT(offs_n);
			GenLD(loadsize);
			mEngine.AddSingle(cmp);
			mEngine.AddJFL(lab_end); // a 'cmp' 0-t ad, ha egyenlo; es ha egyenlo, akkor vege

			mEngine.MakePVT(offs_i);
			GenLD(loadsize);
			if(foreach->IsReverse())
			{
				GenSubImm(loadsize, 1);
			}
			else
			{
				GenAddImm(loadsize, 1);
			}
			mEngine.MakePVT(offs_i);
			GenST(loadsize);
			mEngine.AddJmpL(lab_loop);

			mEngine.AddLabel(lab_end);
		}
		else
		{
			wt_u32 offs_i   = 0;
			wt_u32 offs_n   = 0;
			wt_u32 offs_c   = 0;
			wt_u32 offs_var = 0;
			wt_u32 loadsize = 0;

			if(foreach->GetContainer()->IsSimultaneous() || sizeof(wt_vmptr)!=foreach->GetContainer()->GetType()->GetType(0)->GetLoadSize())
			{
				mCodeGen.AddError(foreach, "Type of the container is not supported.");
				LeaveFrame();
				return false;
			}

			const bool is_seq = wtast::AstNode::SEQTYPE==foreach->GetContainer()->GetType()->GetKind();
// unused?
//			const bool is_tmp = foreach->GetContainer()->IsTemporary();

			if(is_seq)
			{
				loadsize = static_cast<wtast::SeqType*>(foreach->GetContainer()->GetType())->GetElementType()->GetLoadSize();
			}

			wtast::VariableDeclFragment* frag = foreach->GetVariable()->GetFragment(0);

			if(frag->GetVariableDeclaration()->GetType()->IsObject())
			{
				mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), frag->GetName(), frag->GetUniqueId(), offs_var);
			}
			else
			{
				mFrameStack.AddVariable(frag, false, offs_var);
			}

			if(foreach->GetContainer()->IsTemporary())
			{
				mFrameStack.AddVariable(static_cast<wtast::Type*>(foreach->GetContainer()->GetType()), "c", mCodeGen.GetNextUniqueId(), offs_c);
			}
			else
			{
				mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "c", mCodeGen.GetNextUniqueId(), offs_c);
			}

			mFrameStack.AddVariable(mCodeGen.GetIndexTypeU32(), "i", mCodeGen.GetNextUniqueId(), offs_i);

			foreach->GetContainer()->Accept(this);
			mEngine.MakePVT(offs_c);
			mEngine.MakeST(sizeof(wt_vmptr));

			if(foreach->GetContainer()->IsTemporary())
			{
				MakeTDR();
				mEngine.MakePVT(offs_c);
				mCodeGen.GenDestructorMkfp(static_cast<wtast::Type*>(foreach->GetContainer()->GetType()));
				mEngine.AddSingle(WT_TOMB_OP_OSPU);
			}

			std::string lab_loop = mEngine.GenLabel();
			std::string lab_step = mEngine.GenLabel();
			std::string lab_end  = mEngine.GenLabel();

			if(!foreach->IsReverse())
			{
				mFrameStack.AddVariable(mCodeGen.GetIndexTypeU32(), "n", mCodeGen.GetNextUniqueId(), offs_n);

				mEngine.MakePVT(offs_c);
				mEngine.MakeLD(sizeof(wt_vmptr));
				if(is_seq)
				{
					mEngine.AddSingle(WT_TOMB_OP_SEQL);
				}
				else
				{
					mEngine.AddSingle(WT_TOMB_OP_STRL);
				}
				mEngine.MakePVT(offs_n);
				mEngine.MakeST(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
				mEngine.MakePVT(offs_i);
				mEngine.MakeST(sizeof(wt_u32));

				mEngine.AddLabel(lab_loop);
				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.MakePVT(offs_n);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_CMP);
				mEngine.AddSingle(WT_TOMB_OP_QLT);
				mEngine.AddJFL(lab_end);

				mEngine.MakePVT(offs_c);
				mEngine.MakeLD(sizeof(wt_vmptr));
				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));

				if(is_seq)
				{
					mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
					GenLD(loadsize);
					mEngine.MakePVT(offs_var);
					GenST(loadsize);
				}
				else
				{
					mEngine.AddSingle(WT_TOMB_OP_STRCR_S);
					mEngine.AddSingle(WT_TOMB_OP_LDB);
					mEngine.MakePVT(offs_var);
					mEngine.AddSingle(WT_TOMB_OP_STB);
				}

				if(foreach->GetWhile())
				{
					foreach->GetWhile()->Accept(this);
					mEngine.AddJFL(lab_end);
				}

				if(foreach->GetWhich())
				{
					foreach->GetWhich()->Accept(this);
					mEngine.AddJFL(lab_step);
				}

				foreach->GetBody()->Accept(this);

				mEngine.AddLabel(lab_step);

				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
				mEngine.MakePVT(offs_i);
				mEngine.MakeST(sizeof(wt_u32));

				mEngine.AddJmpL(lab_loop);
			}
			else
			{
				mEngine.MakePVT(offs_c);
				mEngine.MakeLD(sizeof(wt_vmptr));
				if(is_seq)
				{
					mEngine.AddSingle(WT_TOMB_OP_SEQL);
				}
				else
				{
					mEngine.AddSingle(WT_TOMB_OP_STRL);
				}
				mEngine.MakePVT(offs_i);
				mEngine.MakeST(sizeof(wt_u32));


				mEngine.AddLabel(lab_loop);
				mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_CMP);
				mEngine.AddSingle(WT_TOMB_OP_QLT);
				mEngine.AddJFL(lab_end);

				mEngine.MakePVT(offs_c);
				mEngine.MakeLD(sizeof(wt_vmptr));
				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);

				if(is_seq)
				{
					mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
					GenLD(loadsize);
					mEngine.MakePVT(offs_var);
					GenST(loadsize);
				}
				else
				{
					mEngine.AddSingle(WT_TOMB_OP_STRCR_S);
					mEngine.AddSingle(WT_TOMB_OP_LDB);
					mEngine.MakePVT(offs_var);
					mEngine.AddSingle(WT_TOMB_OP_STB);
				}

				if(foreach->GetWhile())
				{
					foreach->GetWhile()->Accept(this);
					mEngine.AddJFL(lab_end);
				}

				if(foreach->GetWhich())
				{
					foreach->GetWhich()->Accept(this);
					mEngine.AddJFL(lab_step);
				}

				foreach->GetBody()->Accept(this);

				mEngine.AddLabel(lab_step);

				mEngine.MakePVT(offs_i);
				mEngine.MakeLD(sizeof(wt_u32));
				mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
				mEngine.MakePVT(offs_i);
				mEngine.MakeST(sizeof(wt_u32));

				mEngine.AddJmpL(lab_loop);
			}

			mEngine.AddLabel(lab_end);
		}

		LeaveFrame();
		return false;
	}

	bool Visit(wtast::IfStatement* ifstmt)
	{
		if(ifstmt->GetElse())
		{
			std::string lab_else = mEngine.GenLabel();
			std::string lab_end  = mEngine.GenLabel();

			ifstmt->GetCondition()->Accept(this);
			mEngine.AddJFL(lab_else);
			ifstmt->GetThen()->Accept(this);
			mEngine.AddJmpL(lab_end);
			mEngine.AddLabel(lab_else);
			ifstmt->GetElse()->Accept(this);
			mEngine.AddLabel(lab_end);
		}
		else
		{
			std::string lab_end = mEngine.GenLabel();

			ifstmt->GetCondition()->Accept(this);
			mEngine.AddJFL(lab_end);
			ifstmt->GetThen()->Accept(this);
			mEngine.AddLabel(lab_end);
		}

		return false;
	}

	bool Visit(wtast::AssertStatement* assertion)
	{
		if(mCodeGen.GetOptions()->EnableAsserts())
		{
			if(AddAssert(mCodeGen.mProblemHandler, 0!=mExternals) ||
				AddAssert(mCodeGen.mProblemHandler,
					mExternals->HasException(mCodeGen.AssertionFailExc)))
			{
				return false;
			}

			std::string lab_after = mEngine.GenLabel();

			assertion->GetAssertion()->Accept(this);
			mEngine.AddJTI(lab_after);
			mEngine.MakePBA(mExternals->GetExcAddr(mCodeGen.AssertionFailExc));
			mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
			mEngine.AddSingle(WT_TOMB_OP_ERAISE);
			mEngine.AddLabel(lab_after);
		}

		return false;
	}

/*
	void EndVisit(wtast::AssertStatement*)
	{
	}
*/

	bool Visit(wtast::SeqLiteral* seqlit)
	{
		if(mFrameStack.EnterFrame(""))
		{
			wtast::Expression* init = seqlit->GetInitializer();
			wtparser::UniqueId uniq = mCodeGen.GetNextUniqueId();
			wt_u32             idx  = 0;
			const wt_u32       num  = init ? init->GetType()->GetNumberOfTypes() : 0;
			wt_u32             offs = 0;
			const wt_u32       ldsz = seqlit->GetElementType()->GetLoadSize();
			const bool         obj  = seqlit->GetElementType()->IsObject();

			mEngine.MakePushImm(num);
			mEngine.MakeSEQN(seqlit->GetElementType()->GetLoadSize());
			mEngine.MakeDUP(sizeof(wt_vmptr));
			mEngine.AddMkfp(mCodeGen.GetDestructorName(seqlit->GetSeqType()));
			mEngine.AddSingle(WT_TOMB_OP_TPU);

			if(init)
			{
				mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "", uniq, offs);

				mEngine.MakePVT(offs);
				mEngine.MakeST(sizeof(wt_vmptr));

				for(std::size_t i=0,n=init->GetNumberOfExpressions();i<n;++i)
				{
					wtast::Expression* part = init->GetExpression(i);

					part->Accept(this);

					for(wt_u32 end=idx+part->GetType()->GetNumberOfTypes(),prev=end-1;idx<end;--prev,++idx)
					{
						if(obj)
						{
							if(!part->IsTemporary()) // a 'part'-ot nem lehet tovabbi kifejezesekre bontani
							{
								mCodeGen.GenCopyCall(seqlit->GetElementType());
							}

							MakeTDR();
						}

						mEngine.MakePVT(offs);
						mEngine.MakeLD(sizeof(wt_vmptr));
						mEngine.MakeSEQIR(prev);
						GenST(ldsz);
					}
				}

				// load the created sequence onto datastack
				mEngine.MakePVT(offs);
				mEngine.MakeLD(sizeof(wt_vmptr));
			}

			mFrameStack.LeaveFrame();
		}
		else
		{
			mCodeGen.AddError(seqlit, "Failed to create temporary frame");
		}

		return false;
	}

/*
	void EndVisit(wtast::SeqLiteral*)
	{
	}
*/

	bool Visit(wtast::RecordLiteral* reclit)
	{
		if(mFrameStack.EnterFrame(""))
		{
			wtast::RecordType* rec  = reclit->GetRecordType();
			wtparser::UniqueId uniq = mCodeGen.GetNextUniqueId();
			wt_u32             offs = 0;

			mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "", uniq, offs);

			mEngine.MakeRECN(rec->GetTotalSize());
			mEngine.MakeDUP(sizeof(wt_vmptr));
			mEngine.AddMkfp(mCodeGen.GetDestructorName(rec));
			mEngine.AddSingle(WT_TOMB_OP_TPU);
			mEngine.MakePVT(offs);
			mEngine.MakeST(sizeof(wt_vmptr));

			std::set<std::string> remaining_fields;

			for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
			{
				remaining_fields.insert(rec->GetField(i)->GetName());
			}

			// FieldAssign-bol szarmazo mezok inicializalasa

			for(std::size_t fa_idx=0,fa_num=reclit->GetNumberOfAssignments();fa_idx!=fa_num;++fa_idx)
			{
				wtast::FieldAssign* fa = reclit->GetAssignment(fa_idx);
				wtast::Expression* expr = fa->GetExpression();

				std::size_t field_name_idx = 0;

				for(std::size_t part_idx=0,part_num=expr->GetNumberOfExpressions();part_idx<part_num;++part_idx)
				{
					wtast::Expression* part = expr->GetExpression(part_idx);
					wtast::TypeInfo* part_type = part->GetType();

					part->Accept(this);

					const std::size_t num_types = part_type->GetNumberOfTypes();

					for(std::size_t type_idx=0;type_idx!=num_types;++type_idx)
					{
						const std::string & field_name = fa->GetFieldName(field_name_idx+(num_types-1-type_idx));

						if(AddAssert(mCodeGen.mProblemHandler,
							remaining_fields.count(field_name)))
						{
							return false;
						}

						remaining_fields.erase(field_name);

						wtast::RecordField* field = rec->GetField(field_name);
						wtast::Type* field_type = field->GetType();

						if(field_type->IsObject())
						{
							if(!part->IsTemporary()) // a 'part'-ot nem lehet tovabbi kifejezesekre bontani
							{
								mCodeGen.GenCopyCall(field_type);
							}

							MakeTDR();
						}

						mEngine.MakePVT(offs);
						GenLD(rec->GetLoadSize());
						mEngine.MakeRECFR(field->GetOffset());
						GenST(field_type);
					}

					field_name_idx += num_types;
				}
			}

			// maradek mezok default inicializalasa

			while(!remaining_fields.empty())
			{
				const std::string & field_name = *remaining_fields.begin();
				wtast::RecordField* field = rec->GetField(field_name);
				mCodeGen.DefaultInitializeRecordField(rec, offs, field);
				remaining_fields.erase(remaining_fields.begin());
			}

			// load the created record onto datastack
			mEngine.MakePVT(offs);
			mEngine.MakeLD(sizeof(wt_vmptr));

			mFrameStack.LeaveFrame();
		}
		else
		{
			mCodeGen.AddError(reclit, "Failed to create temporary frame");
		}

		return false;
	}

/*
	void EndVisit(wtast::RecordLiteral*)
	{
	}
*/

	bool Visit(wtast::RecordUpdate* recupd)
	{
		if(mFrameStack.EnterFrame(""))
		{
			wtast::RecordType* rec      = recupd->GetRecordType();
			wtparser::UniqueId uniq     = mCodeGen.GetNextUniqueId();
			wtparser::UniqueId uniq_anc = 0;
			wt_u32             offs_tmp = 0;
			wt_u32             offs_anc = 0;

			mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "", uniq, offs_tmp);

			mEngine.MakeRECN(rec->GetTotalSize());
			mEngine.MakeDUP(sizeof(wt_vmptr));
			mEngine.AddMkfp(mCodeGen.GetDestructorName(rec));
			mEngine.AddSingle(WT_TOMB_OP_TPU);
			mEngine.MakePVT(offs_tmp);
			mEngine.MakeST(sizeof(wt_vmptr));

			std::set<std::string> remaining_fields;

			for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
			{
				remaining_fields.insert(rec->GetField(i)->GetName());
			}

			std::size_t total_assignments = 0;
			for(std::size_t fa_idx=0,fa_num=recupd->GetNumberOfAssignments();fa_idx!=fa_num;++fa_idx)
			{
				total_assignments += recupd->GetAssignment(fa_idx)->GetNumberOfFieldNames();
			}

			if(AddAssert(mCodeGen.mProblemHandler,
				total_assignments<=rec->GetNumberOfFields()))
			{
				return false;
			}

			if(total_assignments!=rec->GetNumberOfFields() ||
				!wtparser::checker::SubtreeIsPure(recupd->GetAncestorExpression()))
			{
				uniq_anc = mCodeGen.GetNextUniqueId();
				mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "", uniq_anc, offs_anc);
				recupd->GetAncestorExpression()->Accept(this);

				mEngine.MakePVT(offs_anc);
				mEngine.MakeST(sizeof(wt_vmptr));
			}

			// FieldAssign-bol szarmazo mezok inicializalasa

			for(std::size_t fa_idx=0,fa_num=recupd->GetNumberOfAssignments();fa_idx!=fa_num;++fa_idx)
			{
				wtast::FieldAssign* fa = recupd->GetAssignment(fa_idx);
				wtast::Expression* expr = fa->GetExpression();

				std::size_t field_name_idx = 0;

				for(std::size_t part_idx=0,part_num=expr->GetNumberOfExpressions();part_idx<part_num;++part_idx)
				{
					wtast::Expression* part = expr->GetExpression(part_idx);
					wtast::TypeInfo* part_type = part->GetType();

					part->Accept(this);

					const std::size_t num_types = part_type->GetNumberOfTypes();

					for(std::size_t type_idx=0;type_idx!=num_types;++type_idx)
					{
						const std::string & field_name = fa->GetFieldName(field_name_idx+(num_types-1-type_idx));

						if(AddAssert(mCodeGen.mProblemHandler,
							remaining_fields.count(field_name)))
						{
							return false;
						}

						remaining_fields.erase(field_name);

						wtast::RecordField* field = rec->GetField(field_name);
						wtast::Type* field_type = field->GetType();

						if(field_type->IsObject())
						{
							if(!part->IsTemporary()) // a 'part'-ot nem lehet tovabbi kifejezesekre bontani
							{
								mCodeGen.GenCopyCall(field_type);
							}

							MakeTDR();
						}

						mEngine.MakePVT(offs_tmp);
						GenLD(rec->GetLoadSize());
						mEngine.MakeRECFR(field->GetOffset());
						GenST(field_type);
					}

					field_name_idx += num_types;
				}
			}

			// copy the remaining fields

			if(AddAssert(mCodeGen.mProblemHandler,
				remaining_fields.empty() || 0!=uniq_anc))
			{
				return false;
			}

			while(!remaining_fields.empty())
			{
				const std::string & field_name = *remaining_fields.begin();
				wtast::RecordField* field = rec->GetField(field_name);
				wtast::Type* field_type = field->GetType();

				mEngine.MakePVT(offs_anc);
				GenLD(rec->GetLoadSize());
				mEngine.MakeRECFR(field->GetOffset());
				GenLD(field_type);

				if(field_type->IsObject())
				{
					mCodeGen.GenCopyCall(field_type);
					MakeTDR();
				}

				mEngine.MakePVT(offs_tmp);
				GenLD(rec->GetLoadSize());
				mEngine.MakeRECFR(field->GetOffset());
				GenST(field_type);

				remaining_fields.erase(remaining_fields.begin());
			}

			if(0!=uniq_anc && recupd->GetAncestorExpression()->IsTemporary())
			{
				mEngine.AddSingle(WT_TOMB_OP_TTD);
			}

			// load the created record onto datastack
			mEngine.MakePVT(offs_tmp);
			mEngine.MakeLD(sizeof(wt_vmptr));

			mFrameStack.LeaveFrame();
		}
		else
		{
			mCodeGen.AddError(recupd, "Failed to create temporary frame");
		}

		return false;
	}

/*
	void EndVisit(wtast::RecordUpdate*)
	{
	}
*/

	bool Visit(wtast::DataTypeLiteral* dtlit)
	{
		if(mFrameStack.EnterFrame(""))
		{
			wtast::DataType*     dt   = dtlit->GetDataType();
			wtast::DataTypeCons* cons = dtlit->GetCons();
			wtparser::UniqueId   uniq = mCodeGen.GetNextUniqueId();
			wt_u32               offs = 0;

			mFrameStack.AddVariable(mCodeGen.GetUnifiedObjectType(), "", uniq, offs);

			mEngine.MakeRECN(cons->GetDataSize());
			mEngine.MakeDUP(sizeof(wt_vmptr));
			mEngine.AddMkfp(mCodeGen.GetDestructorName(dt));
			mEngine.AddSingle(WT_TOMB_OP_TPU);
			mEngine.MakePVT(offs);
			mEngine.MakeST(sizeof(wt_vmptr));

			// set the constructor ID
			mEngine.MakePushImm(cons->GetConsId());
			mEngine.MakePVT(offs);
			mEngine.MakeLD(sizeof(wt_vmptr));
			mEngine.MakeRECFR(0);
			mEngine.MakeST(sizeof(wt_u32));

			// set the refcounter
			mEngine.MakePushImm(static_cast<wt_u32>(1));
			mEngine.MakePVT(offs);
			mEngine.MakeLD(sizeof(wt_vmptr));
			mEngine.MakeRECFR(sizeof(wt_vmptr));
			mEngine.MakeST(sizeof(wt_u32));

			wtast::Expression* init_expr = dtlit->GetInitExpression();

			std::size_t param_base_id = 0;

			if(init_expr)
			{
				for(std::size_t i=0,n=init_expr->GetNumberOfExpressions();i!=n;++i)
				{
					wtast::Expression* expr = init_expr->GetExpression(i);

					expr->Accept(this);

					for(std::size_t j=0,m=expr->GetType()->GetNumberOfTypes();j!=m;++j)
					{
						const std::size_t idx = m-1-j;

						wtast::Type* type = cons->GetType(param_base_id+idx);

						if(type->IsObject())
						{
							if(!expr->IsTemporary())
							{
								mCodeGen.GenCopyCall(type);
							}

							MakeTDR();
						}

						mEngine.MakePVT(offs);
						mEngine.MakeLD(sizeof(wt_vmptr));
						mEngine.MakeRECFR(cons->GetOffset(param_base_id+idx));
						GenST(type);
					}

					param_base_id += expr->GetType()->GetNumberOfTypes();
				}
			}

			// load the created datatype onto datastack
			mEngine.MakePVT(offs);
			mEngine.MakeLD(sizeof(wt_vmptr));

			mFrameStack.LeaveFrame();
		}
		else
		{
			mCodeGen.AddError(dtlit, "Failed to create temporary frame");
		}

		return false;
	}

/*
	void EndVisit(wtast::DataTypeLiteral*)
	{
	}
*/


	bool Visit(wtast::SwitchStatement* sw)
	{
		if(!sw->GetCondition())
		{
			mCodeGen.AddError(sw, "Invalid switch condition");
			return false;
		}

		wtast::TypeInfo* const type_info = sw->GetCondition()->GetType();

		if(!type_info || type_info->IsSimultaneous())
		{
			mCodeGen.AddError(sw, "Invalid type in switch condition");
			return false;
		}

		if(wtast::AstNode::DATATYPE==type_info->GetType(0)->GetKind())
		{
			VisitSwitchDataType(sw, static_cast<wtast::DataType*>(type_info));
		}
		else
		if(wtast::AstNode::ENUMTYPE==type_info->GetType(0)->GetKind())
		{
			VisitSwitchEnumType(sw, static_cast<wtast::EnumType*>(type_info));
		}
		else
		{
			mCodeGen.AddError(sw, "Type in switch condition is not supported");
			return false;
		}

		return false;
	}

/*
	void EndVisit(wtast::SwitchStatement*)
	{
	}
*/
};

class GDtorVisitor : public wtast::AstVisitor
{
	CodeGen &          mCodeGen;
	InstrumentEngine & mEngine;
	FrameStack         mFrameStack;
	bool               mInGlobalBlock;

	void GenPushVarAddr(const FrameStack::Variable* var)
	{
		if(var->is_global)
		{
			mEngine.MakePBA(mCodeGen.GetGlobalAddress() + var->offset);
		}
		else
		{
			mEngine.MakePVT(var->offset);
		}
	}

public:
	GDtorVisitor(CodeGen & cg, InstrumentEngine & eng)
		: wtast::AstVisitor()
		, mCodeGen(cg)
		, mEngine(eng)
		, mInGlobalBlock(false)
	{
		mCodeGen.FillGlobalVariables(mFrameStack);
	}

	bool Visit(wtast::GlobalBlock*)
	{
		mInGlobalBlock = true;
		return true;
	}

	void EndVisit(wtast::GlobalBlock*)
	{
		mInGlobalBlock = false;
	}

	bool Visit(wtast::VariableDeclaration* vardecl)
	{
		if(mInGlobalBlock && !mCodeGen.IsReferredGlobal(vardecl))
		{
			return false;
		}

		return true;
	}

	void EndVisit(wtast::VariableDeclaration* vardecl)
	{
		wtast::Type* type = vardecl->GetType();
		bool type_is_object = type->IsObject();

		if(type_is_object)
		{
			for(std::size_t i=0,n=vardecl->GetNumberOfFragments();i!=n;++i)
			{
				const FrameStack::Variable* var = mFrameStack.FindVar(vardecl->GetFragment(n-i-1)->GetUniqueId());

				GenPushVarAddr(var);
				mEngine.MakeLD(sizeof(wt_vmptr));
				mCodeGen.GenDestructorCall(type);
			}
		}
	}
};

void GenerateCode(CodeGen & cg, InstrumentEngine & eng, DataSection* datasec, ExternalsSection* extsec, wtast::AstNode* astnode, VisitMode visitmode)
{
	if(VISIT_GLOBALDTOR!=visitmode)
	{
		GenVisitor genvisitor(cg, eng, datasec, extsec, visitmode);
		astnode->Accept(&genvisitor);
	}
	else
	{
		GDtorVisitor globaldtor(cg, eng);
		astnode->Accept(&globaldtor);
	}
}

} // codegen
} // wtc
