#include "config_private.h"
#include "hash.h"
#include "byteorder.h"
#include "wtruntime.h"
#include <string.h>
#include "blake2/blake2.h"

#define WT_HASH_MAX_LEN  (512/8)

struct wtr_hashalg
{
	const char* name;
	wt_u16      id;
	wt_u16      len;
};

typedef void (*WTR_PFN_HASH_UPDATE)(void* ctx, const void* data, size_t octets);

static const struct wtr_hashalg wtr_hash_algorithms[] =
{
	{ "BLAKE2-256", WTBHA_BLAKE2_256, 256/8 },
	{ "BLAKE2-384", WTBHA_BLAKE2_384, 384/8 },
	{ "BLAKE2-512", WTBHA_BLAKE2_512, 512/8 },
/*	{ "SHA2-256",   WTBHA_SHA2_256,   256/8 }, */
/*	{ "SHA2-384",   WTBHA_SHA2_384,   384/8 }, */
/*	{ "SHA2-512",   WTBHA_SHA2_512,   512/8 }, */
	{ 0 }
};

wt_u16 wtr_hash_find(const char* name)
{
	const struct wtr_hashalg* alg = wtr_hash_algorithms;

	if(0==strcmp(name, alg->name))
	{
		return alg->id;
	}

	return 0;
}

const char* wtr_hash_name(wt_u16 hashalg)
{
	const struct wtr_hashalg* alg = wtr_hash_algorithms;

	if(alg->id==hashalg)
	{
		return alg->name;
	}

	return NULL;
}

size_t wtr_hash_len(wt_u16 hashalg)
{
	const struct wtr_hashalg* alg = wtr_hash_algorithms;

	if(alg->id==hashalg)
	{
		return alg->len;
	}

	return 0;
}

static wt_error wtr_hash_visitor__(void* hash_ctx, WTR_PFN_HASH_UPDATE update, const void* data, size_t size, wt_bo bo)
{
	const wt_u8* const   base = (const wt_u8*)data;
	const wtb_hdr* const hdr  = (const wtb_hdr*)data;

	if(size<sizeof(wtb_hdr))
	{
		return WTE_BINERR;
	}

	if(WTBMAG0!=hdr->b_ident[WTBI_MAGIC0] ||
		WTBMAG1!=hdr->b_ident[WTBI_MAGIC1] ||
		WTBMAG2!=hdr->b_ident[WTBI_MAGIC2] ||
		WTBMAG3!=hdr->b_ident[WTBI_MAGIC3] ||
		WTBV_CURRENT!=hdr->b_ident[WTBI_VERSION] ||
		bo!=hdr->b_ident[WTBI_BYTEORDER])
	{
		return WTE_BINERR;
	}

	/* header */
	{
		update(hash_ctx, hdr, sizeof(wtb_hdr));
	}

	/* section headers */
	{
		wt_u16 b_shnum = wtr_botoh16(hdr->b_shnum, bo);
		wt_u32 b_shoff = wtr_botoh32(hdr->b_shoff, bo);

		update(hash_ctx, base+b_shoff, sizeof(wtb_shdr)*b_shnum);
	}

	/* thread headers */
	{
		wt_u16 b_thnum = wtr_botoh16(hdr->b_thnum, bo);
		wt_u32 b_thoff = wtr_botoh32(hdr->b_thoff, bo);

		update(hash_ctx, base+b_thoff, sizeof(wtb_thdr)*b_thnum);
	}

	/* section data */
	{
		wt_u16 i;
		wt_u16 b_shnum = wtr_botoh16(hdr->b_shnum, bo);
		wt_u32 b_shoff = wtr_botoh32(hdr->b_shoff, bo);

		const wtb_shdr* shdr = (const wtb_shdr*)(base + b_shoff);

		for(i=1;i<b_shnum;++shdr,++i)
		{
			wt_u32 sh_size = wtr_botoh32(shdr->sh_size, bo);
			wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, bo);

			/* we don't want to hash 0 length data, because it does not any sense */
			if(0==sh_size)
			{
				continue;
			}

			/* CLEAR and LOAD flags excludes themselves */
			if((WTBSF_CLEAR & shdr->sh_flag) && (WTBSF_LOAD & shdr->sh_flag))
			{
				return WTE_BINERR;
			}

			/* if the CLEAR flag is set, it means no associated data */
			if(WTBSF_CLEAR & shdr->sh_flag)
			{
				continue;
			}

			/* INTEGRITY sections: we can hash only the headers, and we should do it */
			/* -- otherwise problems will raise to find the fix-point of the hash value */
			if(WTBST_INTEGRITY==shdr->sh_type)
			{
				update(hash_ctx, base+sh_offs, sizeof(wtb_integhdr));
				continue;
			}

			/* section data is ready to be hashed */
			update(hash_ctx, base+sh_offs, sh_size);
		}
	}

	return WTE_OK;
}

static void wtr_hash_blake2b_update(void* ctx, const void* data, size_t octets)
{
	blake2b_update((blake2b_state*)ctx, data, octets);
}

wt_error wtr_hash_calc(void* out_digest, wt_u16 hashalg, const void* data, size_t size, wt_bo bo)
{
	wt_error err  = WTE_OK;
	size_t   dlen = 256;

	switch(hashalg)
	{
	case WTBHA_BLAKE2_512:
		dlen += 128; /* fall thru' */
	case WTBHA_BLAKE2_384:
		dlen += 128; /* fall thru' */
	case WTBHA_BLAKE2_256:
		{
			blake2b_state blake;

			dlen /= 8;

			blake2b_init(&blake, dlen);

			if(WTE_OK!=(err=wtr_hash_visitor__(&blake, wtr_hash_blake2b_update, data, size, bo)))
			{
				return err;
			}

			blake2b_final(&blake, out_digest, dlen);
		}
		break;

	default:
		return WTE_INTEG;
	}

	return WTE_OK;
}

wt_error wtr_hash_check(const void* digest, wt_u16 hashalg, const void* data, size_t size, wt_bo bo)
{
	wt_u8    calc_digest[WT_HASH_MAX_LEN];
	size_t   hash_len = wtr_hash_len(hashalg);
	wt_error err;

	if(0==hash_len || WT_HASH_MAX_LEN<hash_len)
	{
		return WTE_INTEG;
	}

	if(WTE_OK!=(err=wtr_hash_calc(calc_digest, hashalg, data, size, bo)))
	{
		return err;
	}

	if(0!=memcmp(digest, calc_digest, hash_len))
	{
		return WTE_INTEG;
	}

	return WTE_OK;
}
