#include "config_private.h"
#include "profile.h"
#include "program.h"
#include "stdlib.h"
#include "funcsig.h"
#include "byteorder.h"
#include "function.h"
#include "loader.h"
#include "processor.h"
#include "routines.h"
#include "type.h"
#include "operator.h"
#include "vector.h"
#include "ctx.h"
#include <string.h>
#include <stdlib.h>

#define VMMA_BLOCKSIZE     4096
#define VMMA_CLIENTALIGN     64
#define EXCVEC_CAPA          16
#define TYPEVEC_CAPA         16

/*----------------------------------------------------------------*/

void* wtr_profile_vm_malloc(struct wtr_profile* prof, size_t size)
{
	return vmma_alloc(prof->alloc_client, size);
}

void wtr_profile_vm_free(struct wtr_profile* prof, void* memory)
{
	vmma_free(prof->alloc_client, memory);
}

void* wtr_profile_sys_realloc(struct wtr_profile* const prof, void* memory,
	const size_t old_size, const size_t new_size)
{
	void* const  alloc_ctx  = prof->alloc_ctx;
	WT_PFN_ALLOC alloc_func = prof->alloc_func;
	void*        new_mem;

	if(NULL==memory)
	{
		return alloc_func(alloc_ctx, NULL, new_size);
	}

	if(old_size==new_size)
	{
		/* this should not happen, but since they are equal do nothing */
		return memory;
	}

	if(NULL!=(new_mem=alloc_func(alloc_ctx, memory, new_size)))
	{
		return new_mem;
	}

	/* alloc_func() could not reallocate the block, but this could be */
	/* because it does not implement realloc, so try a different approach */

	if(NULL==(new_mem=alloc_func(alloc_ctx, NULL, new_size)))
	{
		/* nope, we sure ran out of memory */
		return NULL;
	}

	if(old_size>new_size)
	{
		memcpy(new_mem, memory, new_size);
	}
	else
	{
		memcpy(new_mem, memory, old_size);
	}

	alloc_func(alloc_ctx, memory, 0);

	return new_mem;
}

void* wtr_profile_sys_malloc(struct wtr_profile* const prof, size_t size)
{
	return prof->alloc_func(prof->alloc_ctx, NULL, size);
}

void wtr_profile_sys_free(struct wtr_profile* const prof, void* memory)
{
	prof->alloc_func(prof->alloc_ctx, memory, 0);
}

struct wtr_exception* wtr_profile_makeexception(struct wtr_profile* prof, struct wtr_program* prog, const char* str)
{
	const size_t                len = strlen(str) + 1;
	struct wtr_exception* const exc = (struct wtr_exception*)wtr_profile_sys_malloc(prof, sizeof(struct wtr_exception)+len*sizeof(*str));

	if(NULL!=exc)
	{
		exc->refcount  = 1;
		exc->exception = 0;
		exc->owner     = prog;
		memcpy(exc->name, str, len*sizeof(*str));
	}

	return exc;
}

struct wtr_exception* wtr_profile_findexception(struct wtr_profile* const vm,
	const char* const excname)
{
	const size_t size = wtr_vector_size(&vm->exception_vec);

	if(0<size)
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&vm->exception_vec, 0);

		struct wtr_exception** const end = ptr + size;

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;

			if(NULL!=exc && 0==strcmp(exc->name, excname))
			{
				return exc;
			}
		}
	}

	return NULL;
}

wt_bool wtr_profile_removeexception(struct wtr_profile* const vm,
	struct wtr_exception* const exc)
{
	size_t size = wtr_vector_size(&vm->exception_vec);

	if(0<exc->refcount || exc->exception>=size)
	{
		return WT_FALSE;
	}

	wtr_vector_set(&vm->exception_vec, exc->exception, NULL);

	memset(exc, 0, sizeof(struct wtr_exception));
	wtr_profile_sys_free(vm, exc);

	/* pop trailing NULLs, but the vector must have 1 remaining element */

	while(1<(size=wtr_vector_size(&vm->exception_vec)) &&
		NULL==wtr_vector_at(&vm->exception_vec, size-1))
	{
		wtr_vector_pop_back(&vm->exception_vec);
	}

	return WT_TRUE;
}

wt_error wtr_profile_addexception(struct wtr_profile* const vm,
	struct wtr_program* const prog, const char* const excname,
	struct wtr_exception** const p_exc)
{
	struct wtr_exception* exception;

	if(NULL!=(exception=wtr_profile_findexception(vm, excname)))
	{
		*p_exc = exception;
		++exception->refcount;
		return WTE_OK;
	}
	else
	{
		if(NULL==(exception=wtr_profile_makeexception(vm, prog, excname)))
		{
			return WTE_NOMEM;
		}
		else
		{
			wt_exception exc_id = 0;
			char*        name   = exception->name;
			wt_bool      push   = WT_TRUE;
			const size_t size   = wtr_vector_size(&vm->exception_vec);

			for(;exc_id<size;++exc_id)
			{
				if(NULL==wtr_vector_at(&vm->exception_vec, exc_id))
				{
					push = WT_FALSE;
					break;
				}
			}

			if(
				(push && 0!=wtr_vector_push_back(&vm->exception_vec, vm,
					exception))
				||
				(!push && 0!=wtr_vector_set(&vm->exception_vec, exc_id,
					exception))
			)
			{
				wtr_profile_sys_free(vm, exception);
				return WTE_NOMEM;
			}

			*p_exc = exception;
			exception->exception = exc_id;
		}
	}

	return WTE_OK;
}

struct wtr_function* wtr_profile_findfunction(struct wtr_profile* const vm,
	struct wtr_funcsig* const sig)
{
	struct wtr_list* const list = &vm->list_functions;
	struct wtr_node*       it;

	wtr_list_for_each(list, it)
	{
		struct wtr_function* const func = wtr_contof(it, struct wtr_function,
			link_functions);

		if(0==wtr_funcsig_cmp(sig, func->funcsig))
		{
			return func;
		}
	}

	return NULL;
}

struct wtr_type* wtr_profile_findtype(struct wtr_profile* const vm,
	const char* const name)
{
	const size_t size = wtr_vector_size(&vm->type_vec);

	if(0<size)
	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&vm->type_vec, 0);

		struct wtr_type** const end = ptr + size;

		while(ptr<end)
		{
			struct wtr_type* const typ = *ptr++;

			if(NULL!=typ && 0==strcmp(typ->name, name))
			{
				return typ;
			}
		}
	}

	return NULL;
}

wt_bool wtr_profile_addtype(struct wtr_profile* const vm,
	struct wtr_type* const type)
{
	wt_u32 idx = 0;
	wt_u32 num = wtr_vector_size(&vm->type_vec);

	/* search an unused index */

	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&vm->type_vec, 0);

		for(idx=0;idx!=num;++idx)
		{
			if(NULL==(*ptr))
			{
				break;
			}

			++ptr;
		}
	}

	type->id = idx;

	if(idx==num)
	{
		if(0!=wtr_vector_push_back(&vm->type_vec, vm, type))
		{
			return WT_FALSE;
		}
	}
	else
	{
		if(0!=wtr_vector_set(&vm->type_vec, idx, type))
		{
			return WT_FALSE;
		}
	}

	return WT_TRUE;
}

void wtr_profile_removetype(struct wtr_profile* const vm,
	struct wtr_type* const type)
{
	wtr_vector_set(&vm->type_vec, type->id, NULL);
	wtr_type_release(type);
}

struct wtr_operator* wtr_profile_findoperator(struct wtr_profile* const vm,
	const char* const name)
{
	struct wtr_list* const list = &vm->list_operators;
	struct wtr_node*       it;

	wtr_list_for_each(list, it)
	{
		struct wtr_operator* const op = wtr_contof(it, struct wtr_operator,
			link_ops);

		if(0==strcmp(name, op->name))
		{
			return op;
		}
	}

	return NULL;
}

wt_bool wtr_profile_addoperator(struct wtr_profile* const vm,
	struct wtr_operator* const op)
{
	if(NULL!=wtr_profile_findoperator(vm, op->name))
	{
		return WT_FALSE;
	}

	wtr_list_insert_tail(&vm->list_operators, &op->link_ops);

	return WT_TRUE;
}

void wtr_profile_removeoperator(struct wtr_profile* const vm,
	struct wtr_operator* const op)
{
	wtr_list_remove(&op->link_ops);
	wtr_list_init(&op->link_ops);
	wtr_operator_release(op);
}

size_t wtr_sizeof(wt_u32 type)
{
	if(type & (WTTF_SEQ | WTTF_REF | WTTF_LOCAL))
	{
		return sizeof(wt_vmptr);
	}

	switch(type & WTTM_BASICTYPE)
	{
	case WTT_BOOL:    return sizeof(wt_bool);
	case WTT_STRING:  return sizeof(wt_vmptr);  /* this is an object */
	case WTT_I32:     return sizeof(wt_i32);
	case WTT_U32:     return sizeof(wt_u32);
	case WTT_I64:     return sizeof(wt_i64);
	case WTT_U64:     return sizeof(wt_u64);
#if WTRUNTIME_ENABLE_FLOAT32
	case WTT_F32:     return sizeof(wt_f32);
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
	case WTT_F64:     return sizeof(wt_f64);
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
	case WTT_CHAR:    return sizeof(wt_char);
	case WTT_HANDLE:  return sizeof(wt_handle);
	case WTT_FILE:    return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTT_DIR:     return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTT_PROGRAM: return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTT_I8:
	case WTT_U8:
	case WTT_I16:
	case WTT_U16:     return sizeof(wt_u32);
	}

	return 0;
}

/*---------------------------------------------------------------- */

wt_error wtrCreateTombVM(wt_profile* const p_profile,
	wt_u32 ver, wt_u32 flags, void* memdata, wt_u32 memsize, wt_u8 ratio_css,
	wt_u8 ratio_dss, void* alloc_ctx, WT_PFN_ALLOC alloc_func)
{
	void*                allocated = NULL;
	wt_u8*               memory    = NULL;
	struct wtr_profile*  profile   = NULL;
	struct vmma_context* a_ctx     = NULL;
	struct vmma_client*  a_client  = NULL;
	wt_u32               acc_flags = 0; /* accepted flags (WTRVMF_*) */
	void*                pointer;

	if(NULL==p_profile || 0==ratio_css || 0==ratio_dss || NULL==alloc_func)
	{
		return WTE_BADARG;
	}

	if(memsize<VMMA_BLOCKSIZE*8)
	{
		return WTE_MEMSZ;
	}

	if(NULL==memdata)
	{
		memdata = alloc_func(alloc_ctx, NULL, memsize);

		if(NULL==memdata)
		{
			return WTE_NOMEM;
		}

		allocated = memdata;
	}

	memory = (wt_u8*)memdata;

	if(flags & WTRVMF_REQINTEGRITY)
	{
		acc_flags |= WTRVMF_REQINTEGRITY;
	}

	if(NULL==(a_ctx=vmma_create_context(memory, memsize, VMMA_CLIENTALIGN,
		VMMA_BLOCKSIZE, ((2048*1024)>=memsize) ? VMMA_FL_MINGROWTH : 0)))
	{
		goto error_1;
	}

	if(NULL==(a_client=vmma_create_client(a_ctx, 0)))
	{
		goto error_2;
	}

	if(NULL==(profile=(struct wtr_profile*)vmma_alloc(a_client, sizeof(struct wtr_profile))))
	{
		goto error_2;
	}

	memset(profile, 0, sizeof(struct wtr_profile));

	profile->alloc_ctx      = alloc_ctx;
	profile->alloc_func     = alloc_func;
	profile->ratio_css      = ratio_css;
	profile->ratio_dss      = ratio_dss;
	profile->alloc_context  = a_ctx;
	profile->alloc_client   = a_client;
	profile->memory         = memory;
	profile->total_memsize  = memsize;
	profile->allocated_mem  = allocated;
	profile->accepted_flags = acc_flags;
	profile->xs_x           = (wt_u32)(wt_uintptr)profile;
	profile->xs_y           = (wt_u32)(wt_uintptr)wtrCreateTombVM;
	profile->xs_z           = 0x5a5a5a5a ^ profile->xs_x ^ profile->xs_y;
	profile->xs_w           = 0xa5a5a5a5 ^ profile->xs_x ^ profile->xs_y;

	wtr_list_init(&profile->list_programs);
	wtr_list_init(&profile->list_functions);
	wtr_list_init(&profile->list_operators);

	if(0!=wtr_vector_create(&profile->exception_vec, profile, EXCVEC_CAPA))
	{
		goto error_2;
	}

	if(0!=wtr_vector_create(&profile->type_vec, profile, TYPEVEC_CAPA))
	{
		goto error_2;
	}

	if(NULL==(pointer=vmma_alloc(a_client, sizeof(wtr_routine_tcl_ja))))
	{
		goto error_2;
	}
	else
	{
		profile->pc_tcl_ja = ((wt_u16*)pointer) - ((wt_u16*)memory);
		memcpy(pointer, wtr_routine_tcl_ja, sizeof(wtr_routine_tcl_ja));
	}

	if(NULL==(pointer=vmma_alloc(a_client, sizeof(wtr_routine_tcl_leave))))
	{
		goto error_2;
	}
	else
	{
		profile->pc_tcl_leave = ((wt_u16*)pointer) - ((wt_u16*)memory);
		memcpy(pointer, wtr_routine_tcl_leave, sizeof(wtr_routine_tcl_leave));
	}

	if(NULL==(pointer=vmma_alloc(a_client, sizeof(wtr_routine_strd_leave))))
	{
		goto error_2;
	}
	else
	{
		profile->pc_strd_leave = ((wt_u16*)pointer) - ((wt_u16*)memory);
		memcpy(pointer, wtr_routine_strd_leave, sizeof(wtr_routine_strd_leave));
	}

	if(NULL==(pointer=vmma_alloc(a_client, sizeof(wtr_routine_cfp_leave))))
	{
		goto error_2;
	}
	else
	{
		profile->pc_cfp_leave = ((wt_u16*)pointer) - ((wt_u16*)memory);
		memcpy(pointer, wtr_routine_cfp_leave, sizeof(wtr_routine_cfp_leave));
	}

	profile->external_tag = ~((wt_u32)0);

	{
		struct wtr_exception* ex;

#define WTR_PROFILE_ADDEXCEPTION(exc) \
		if(WTE_OK!=wtr_profile_addexception(profile, NULL, #exc, &ex)) \
		{ \
			goto error_2; \
		}

		WTR_STDEXCEPTIONS(WTR_PROFILE_ADDEXCEPTION);
	}

	*p_profile = profile;
	return WTE_OK;

error_2:;
	vmma_delete_context(a_ctx);
error_1:
	if(allocated) { alloc_func(alloc_ctx, allocated, 0); }
	return WTE_NOMEM;
}

wt_error wtrProfileDestroy(wt_profile profile)
{
	size_t size;

	if(NULL==profile)
	{
		return WTE_BADARG;
	}

	if(0<profile->num_context)
	{
		return WTE_REFCTX;
	}

	while(0<(size=wtr_vector_size(&profile->exception_vec)))
	{
		struct wtr_exception* exc;

		struct wtr_exception** const p_exc =
			(struct wtr_exception**)wtr_vector_at(&profile->exception_vec,
				size-1);

		if(NULL==p_exc || NULL==(exc=(*p_exc)))
		{
			wtr_vector_pop_back(&profile->exception_vec);
			continue;
		}

		if(NULL!=exc->owner || 1!=exc->refcount)
		{
			return WTE_REFER;
		}

		--exc->refcount;

		if(!wtr_profile_removeexception(profile, exc))
		{
			return WTE_REFER;
		}
	}

	wtr_vector_delete(&profile->exception_vec, profile);
	wtr_vector_delete(&profile->type_vec, profile);

	vmma_delete_context(profile->alloc_context);

	if(NULL!=profile->allocated_mem)
	{
		WT_PFN_ALLOC alloc_func = profile->alloc_func;
		void*        alloc_ctx  = profile->alloc_ctx;

		alloc_func(alloc_ctx, profile->allocated_mem, 0);
	}

	return WTE_OK;
}

wt_error wtrProfileSetLoaderOps(wt_profile profile, void* const ops_ctx,
	const struct wtr_loaderops* const ops)
{
	profile->loaderops_ctx = ops_ctx;
	profile->loaderops     = ops;

	return WTE_OK;
}

wt_error wtrProfileGetException(wt_profile profile, const char* const name,
	wt_exception* const p_exc)
{
	if(NULL==name || NULL==p_exc)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_exception* const exception =
			wtr_profile_findexception(profile, name);

		if(NULL==exception)
		{
			return WTE_NOENT;
		}

		*p_exc = exception->exception;
	}

	return WTE_OK;
}

wt_error wtrProfileExceptionToString(wt_profile profile, wt_exception exc,
	char* namebuf, size_t bufsize)
{
	if(NULL==namebuf)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_exception** const p_exc =
			(struct wtr_exception**)wtr_vector_at(&profile->exception_vec,
			exc);

		size_t len;

		if(NULL==p_exc)
		{
			return WTE_NOENT;
		}

		len = strlen((*p_exc)->name);

		if(len>=bufsize)
		{
			return WTE_NOFIT;
		}

		memcpy(namebuf, (*p_exc)->name, len+1);
	}

	return WTE_OK;
}

wt_error wtrProfileAddException(wt_profile profile, const char* const name,
	wt_exception* const out_exc)
{
	struct wtr_exception* ex;
	const wt_error err = wtr_profile_addexception(profile, NULL, name, &ex);

	if(WTE_OK==err && NULL!=out_exc)
	{
		*out_exc = ex->exception;
	}

	return err;
}

wt_error wtrProfileMemoryStatus(wt_profile profile, wt_u64* p_total, wt_u64* p_free)
{
	wt_error err;

	if(!p_total || !p_free)
	{
		err = WTE_BADARG;
	}
	else
	{
		*p_total = vmma_get_total_size(profile->alloc_context);
		*p_free  = vmma_get_free_size (profile->alloc_context);
		err      = WTE_OK;
	}

	return err;
}

wt_error wtrProfileAddFunction(wt_profile profile, const char* name,
	wt_u32 argc, const wt_u32* argv, wt_u32 retc, const wt_u32* retv,
	WT_PFN_EXTFUNC func, void* token)
{
	if(NULL==name || NULL==func || !(argc==0 || argv) || !(retc==0 || retv))
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_funcsig* const sig = wtr_funcsig_new(profile, name, argc,
			argv, retc, retv, WT_FALSE, NULL, wtr_byteorder());

		if(NULL==sig)
		{
			return WTE_NOMEM;
		}

		if(NULL!=wtr_profile_findfunction(profile, sig))
		{
			wtr_funcsig_delete(profile, sig);
			return WTE_DUPENT;
		}
		else
		{
			struct wtr_cfunction* const cf = wtr_cfunction_new(profile, sig,
				func, token);

			if(NULL==cf)
			{
				wtr_funcsig_delete(profile, sig);
				return WTE_NOMEM;
			}
			else
			{
				struct wtr_function* const fh = wtr_function_new(profile, sig,
					cf, NULL, 0);

				if(NULL==fh)
				{
					wtr_cfunction_delete(profile, cf);
					wtr_funcsig_delete(profile, sig);
					return WTE_NOMEM;
				}
			}
		}
	}

	return WTE_OK;
}

wt_error wtrProfileGetFunction(wt_profile profile, wt_function* p_func,
	const char* name, wt_u32 argc, const wt_u32* argv, wt_u32 retc,
	const wt_u32* retv, wt_bool pure)
{
	if(NULL==p_func || NULL==name)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_funcsig* const sig = wtr_funcsig_new(profile, name, argc,
			argv, retc, retv, pure, NULL, wtr_byteorder());

		if(NULL==sig)
		{
			return WTE_NOMEM;
		}
		else
		{
			struct wtr_function* const func =
				wtr_profile_findfunction(profile, sig);

			if(NULL==func)
			{
				return WTE_NOENT;
			}
			else
			{
				*p_func = wtr_function_genhandle(profile, func);
				wtr_funcsig_delete(profile, sig);

				if(NULL==(*p_func))
				{
					return WTE_NOMEM;
				}
			}
		}
	}

	return WTE_OK;
}

wt_error wtrProfileReleaseFunction(wt_profile profile, wt_function func)
{
	if(NULL==func)
	{
		return WTE_BADARG;
	}

	wtr_function_remhandle(profile, func);
	return WTE_OK;
}

wt_error wtrProfileLoadProgram(wt_profile profile, wt_program* program,
	struct wtr_context* const ctx, const void* data, size_t size,
	wt_u32 flags)
{
	if(NULL==program || NULL==data || 0==size || NULL==ctx)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	return wtr_loader(profile, program, ctx, data, size, flags);
}

wt_error wtrProfileReleaseProgram(wt_profile profile, wt_program program,
	struct wtr_context* const ctx)
{
	size_t i;

	if(NULL==profile || NULL==program || NULL==ctx ||
		program->profile!=profile)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	if(1<program->refcount)
	{
		return WTE_REFER;
	}
	else
	if(0==program->refcount)
	{
		return WTE_DESTRP;
	}

	if(0<wtr_vector_size(&program->own_exceptions))
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&program->own_exceptions, 0);

		struct wtr_exception** const end = ptr +
			wtr_vector_size(&program->own_exceptions);

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;

			/* the refcount can be at most 1 (reference counter will be */
			/* increased by 1 when the program imports it) */

			if(1<exc->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	if(0<wtr_vector_size(&program->own_types))
	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&program->own_types, 0);

		struct wtr_type** const end = ptr +
			wtr_vector_size(&program->own_types);

		while(ptr<end)
		{
			struct wtr_type* const typ = *ptr++;

			/* the refcount can be at most 1 (reference counter will be */
			/* increased by 1 when the program imports it) */

			if(1<typ->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	if(0<wtr_vector_size(&program->own_operators))
	{
		struct wtr_operator** ptr = (struct wtr_operator**)
			wtr_vector_at(&program->own_operators, 0);

		struct wtr_operator** const end = ptr +
			wtr_vector_size(&program->own_operators);

		while(ptr<end)
		{
			struct wtr_operator* const op = *ptr++;

			/* the refcount shall be zero now */

			if(0<op->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	--program->refcount;

	if(program->pc_dtor)
	{
		wtr_ctx_invoke_dtor(ctx, program);
	}
	else
	{
		wtr_program_destroy(profile, program);
	}

	return WTE_OK;
}

wt_error wtrProfileSetUserPointer(wt_profile profile, void* ptr)
{
	profile->user_ptr = ptr;
	return WTE_OK;
}

wt_error wtrProfileGetUserPointer(wt_profile profile, void** pptr)
{
	*pptr = profile->user_ptr;
	return WTE_OK;
}
